# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
import scipy as sp
from oedes.functions import wavelength_to_photon_energy
from oedes.optical.func import wavelength_range
import scipy.interpolate

__all__ = ['IsotropicLayer', 'NormalIncidence']


def forward(n, angle):
    "Decide if angle is forward traveling plane wave in medium with index n"
    if not np.all(np.real(n) * np.imag(n) >= 0):
        raise ValueError('amplification of light not supported')

    # Decision on basis of direction of Poynting vector or lack of gain
    x = n * np.cos(angle)
    return np.where(np.abs(x.imag) > 0, x.imag > 0, x.real > 0)


def snell_forward(n, angle_of_incidence):
    """
    return list of angle theta in each layer based on angle th_0 in layer 0,
    using Snell's law
    """
    angles = sp.arcsin(n[0]*np.sin(angle_of_incidence) / n)
    return np.where(forward(n, angles), angles, np.pi-angles)


def _aux_calculations(polarization_is_s, n_i, n_f, th_i, th_f):
    c_i = np.cos(th_i)
    c_f = np.cos(th_f)
    k_i = n_i * c_i
    k_f = n_f * c_f
    return dict(c_i=c_i, c_f=c_f, k_i=k_i, k_f=k_f)


def _power_calculation(r, t, polarization_is_s, n_i, n_f, aux):
    k_i_c = n_i * np.conj(aux['c_i'])
    k_f_c = n_f * np.conj(aux['c_f'])
    aux.update(k_i_c=k_i_c, k_f_c=k_f_c)
    r_ = np.abs(r)
    R = r_*r_
    t_ = np.abs(t)
    T = t_*t_ * np.where(polarization_is_s,
                         aux['k_f'].real / aux['k_i'].real,
                         k_f_c.real / k_i_c.real)
    return R, T


class FresnelCalculation(object):
    def __init__(self, polarization_is_s, n_i, n_f,
                 th_i, th_f, with_power=False):
        def _do(c_i, c_f, k_i, k_f):
            u_i = n_i * c_f
            u_f = n_f * c_i
            self.r = np.where(polarization_is_s,
                              (k_i - k_f) / (k_i + k_f),
                              (u_f - u_i) / (u_f + u_i))
            self.t = 2 * k_i / \
                np.where(polarization_is_s, k_i + k_f, u_f + u_i)
        self.aux = _aux_calculations(polarization_is_s, n_i, n_f, th_i, th_f)
        _do(**self.aux)
        if with_power:
            self.R, self.T = _power_calculation(
                self.r, self.t, polarization_is_s, n_i, n_f, self.aux)
            if with_power:
                self.R, self.T = _power_calculation(
                    self.r, self.t, polarization_is_s, n_i, n_f, self.aux)


class PowerCalculation(object):
    def __init__(self, r, t, polarization_is_s, n_i, n_f, th_i, th_f):
        aux = _aux_calculations(polarization_is_s, n_i, n_f, th_i, th_f)
        self.R, self.T = _power_calculation(
            r, t, polarization_is_s, n_i, n_f, aux)
        k_i, k_i_c = aux['k_i'], aux['k_i_c']
        self.power_entering = np.where(polarization_is_s,
                                       (k_i*(1+np.conj(r))*(1-r)).real / k_i.real,
                                       (k_i_c*(1.+r)*(1.-np.conj(r))).real / k_i_c.real)

class AbsorptionCalculation(object):
    def __init__(self, kz, z, vw, polarization_is_s, n, n0, theta, theta0, all_values=True):
        # Amplitude of forward-moving wave is Ef, backwards is Eb
        Ef = vw[0] * np.exp(1j * kz * z)
        Eb = vw[1] * np.exp(-1j * kz * z)
        c0 = np.cos(theta0)
        c = np.cos(theta)
        ds = 1. / (n0*c0).real
        dp = 1. / (n0*np.conj(c0)).real
        self.poyn = np.where(polarization_is_s, (n*c*np.conj(Ef+Eb)*(Ef-Eb)).real * ds,
                        (n*np.conj(c)*(Ef+Eb)*np.conj(Ef-Eb)).real * dp)
        if not all_values:
            return
        p_ = abs(Ef+Eb)
        q_ = abs(Ef-Eb)
        p = p_*p_
        q = q_*q_
        self.absor = np.where(polarization_is_s, (n*c*kz*p).imag *
                         ds, (n*np.conj(c)*(kz*q-np.conj(kz)*p)).imag * dp)
        Ex = np.where(polarization_is_s, 0, (Ef - Eb) * c)
        Ey = np.where(polarization_is_s, Ef + Eb, 0)
        Ez = np.where(polarization_is_s, 0, (-Ef - Eb) * np.sin(theta))
        self.E = (Ex, Ey, Ez)
        self.z = z


def mat2x2(a, b, c, d):
    return ((a, b), (c, d))


def dot(X, Y):
    (a1, b1), (c1, d1) = X
    (a2, b2), (c2, d2) = Y
    return mat2x2(a1*a2+b1*c2, a1*b2+b1*d2, c1*a2+d1*c2, c1*b2+d1*d2)


def dotvec(M, x):
    (a, b), (c, d) = M
    v, w = x
    return (a*v+b*w, c*v+d*w)


def transfer_matrix(thickness, n, polarization_is_s, theta_0, wavelength):
    if thickness[0] != np.inf or thickness[-1] != np.inf:
        raise ValueError('first and last layer should be infinite')
    thickness = thickness.copy()
    thickness[0] = -1.
    thickness[-1] = -1.

    kx = n[0]*np.sin(theta_0)
    if not np.all(kx.imag <= 100*np.abs(kx)*np.finfo(kx.dtype).eps):
        raise ValueError('n*theta[0] must be real in first layer')
    if not np.all(forward(n[0], theta_0)):
        raise ValueError('theta0 must be forward')
    theta = snell_forward(n, theta_0)
    kz = 2 * np.pi * n * np.cos(theta) / wavelength
    delta = kz * thickness[..., np.newaxis]
    num_layers = len(thickness)

    def fresnel_matrix(i):
        f = FresnelCalculation(polarization_is_s, n[i], n[i+1],
                               theta[i], theta[i+1])
        u = 1./f.t
        return mat2x2(u, u*f.r, u*f.r, u)

    def propagation_matrix(i):
        return mat2x2(np.exp(-1j*delta[i]), 0, 0, np.exp(1j*delta[i]))

    matrices = [fresnel_matrix(
        0)]+[dot(propagation_matrix(i), fresnel_matrix(i)) for i in range(1, num_layers-1)]
    Mfull = matrices[0]
    for M in matrices[1:]:
        Mfull = dot(Mfull, M)

    # Net complex transmission and reflection amplitudes
    (a, b), (c, d) = Mfull
    t = 1./a
    r = c*t

    amplitudes = [(t, np.zeros_like(t))]
    for i in range(num_layers-2, 0, -1):
        amplitudes.append(dotvec(matrices[i], amplitudes[-1]))
    amplitudes.append(np.zeros_like(np.asarray(amplitudes[0])))
    vw_list = np.asarray(list(map(np.asarray, amplitudes[::-1])))

    p = PowerCalculation(r, t, polarization_is_s,
                         n[0], n[-1], theta[0], theta[-1])

    return dict(r=r, t=t, M_list=matrices, th_list=theta, kz_list=kz, num_layers=num_layers,
                vw_list=vw_list, R=p.R, T=p.T, power_entering=p.power_entering)


class IsotropicLayer(object):
    def __init__(self, thickness, material):
        self.thickness = thickness
        self.material = material

    def refractive_index(self, wavelength):
        return self.material.refractive_index(wavelength)


class TransferMatrixResult:
    def __init__(self, layers, polarization_is_s, theta_0, wavelength):
        """
        Run transfer matrix calculation, with layers, and with (polarization_is_s, theta_incidence, wavelength) broadcast against each other
        Layer before first layer, and layer after last is vacuum/air with n=1
        """
        thickness = np.asarray(
            [np.inf] + [l.thickness for l in layers] + [np.inf])
        #polarization_is_s, wavelength, theta_0 = np.broadcast_arrays(polarization_is_s, wavelength, theta_0)
        n = np.asarray([np.ones_like(wavelength)] +
                       [l.refractive_index(wavelength) for l in layers] +
                       [np.ones_like(wavelength)])
        self.result = transfer_matrix(
            thickness, n, polarization_is_s, theta_0, wavelength)
        self.layer_dict = dict((id(layer), i+1)
                               for i, layer in enumerate(layers))
        self.n = n
        self.polarization_is_s = polarization_is_s
        a = AbsorptionCalculation(self.result['kz_list'], 0., (self.result['vw_list'][:,0,:],self.result['vw_list'][:,1,:]), polarization_is_s, n, n[0], self.result['th_list'], self.result['th_list'][0], all_values=False)
        self.incident_power_by_layer = a.poyn[1:]
        self.T = self.result['T']
        self.R = self.result['R']


    def absorption(self, layer, x):
        assert np.all(x >= 0)
        assert np.all(x <= layer.thickness)
        layer_number = self.layer_dict[id(layer)]
        th_list = self.result['th_list']
        return AbsorptionCalculation(self.result['kz_list'][layer_number], x[:, np.newaxis], self.result['vw_list'][layer_number],
                                     self.polarization_is_s, self.n[layer_number], self.n[0], th_list[layer_number], th_list[0])

    def layer_index(self, layer):
        return self.layer_dict[id(layer)]


class NormalIncidenceResultForLayer(object):
    def __init__(self, ni, layer, x):
        self.wavelengths = ni.wavelengths
        self.density = ni.density
        self.tmm = ni.tmm
        self.result = self.tmm.absorption(layer, x)
        self.layer = layer
        self._integrate = ni.integrate
        self.x = x

    def light_power(self):
        return self._integrate(self.result.poyn)

    def absorption(self):
        return self._integrate(self.result.absor)

    def photons_absorbed(self):
        return self._integrate(
            self.result.absor/wavelength_to_photon_energy(self.wavelengths))

    def interpolated(self, y):
        return scipy.interpolate.interp1d(
            self.x, y, kind='linear', bounds_error=True)

    absorption_in_photons = photons_absorbed


class NormalIncidence:
    def __init__(self, layers, spectrum, wavelengths, clip_wavelengths=True):
        if clip_wavelengths:
            a, b = wavelength_range(
                spectrum, *(layer.material.refractive_index for layer in layers))
            wavelengths = wavelengths[(wavelengths >= a) & (wavelengths <= b)]
        self.wavelengths = wavelengths
        self.density = spectrum(wavelengths)
        self.tmm = TransferMatrixResult(layers, True, 0, wavelengths)
        self.layers = layers
        self.incident_power_by_layer = self.integrate(self.tmm.incident_power_by_layer)
        self.T = self.integrate(self.tmm.T)
        self.R = self.integrate(self.tmm.R)

    def in_layer(self, layer, x):
        return NormalIncidenceResultForLayer(self, layer, x)

    def total_absorption_in_layer(self, layer):
        i = self.tmm.layer_index(layer)
        return self.incident_power_by_layer[i-1]-self.incident_power_by_layer[i]

    def integrate(self, y, axis=None):
        return np.trapz(y*self.density, self.wavelengths, axis=y.ndim-1)

    def illuminating_power(self):
        return self.integrate(np.asarray(1))



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# This tests if diffusion equation is solved correctly with different
# boundary conditions

import matplotlib.pylab as plt
from oedes.testing import store

make_plots = False

from oedes import solve, bdf1adapt, models, transientsolve
from oedes.fvm import mesh1d
from oedes.models import AppliedVoltage, Poisson, AdvectionDiffusion
from oedes import context
from oedes import testing
import numpy as np
import unittest
from oedes.models import Zero, Equal
from oedes import mpl


def integrate_c_internal(eq, c):
    return np.sum((eq.mesh.cells['volume'] * c)[eq.mesh.internal.idx])


def integrate_c_all(eq, c):
    return np.sum((eq.mesh.cells['volume'] * c))


def run_diffusion(L, times, bc_equation=None,
                  integrate_c=integrate_c_internal, rtol=1e-3, atol=0., return_context=False, **kwargs):
    b = models.BaseModel()
    mesh = mesh1d(L, dx_boundary=L * 1e-5)
    # Poisson's equation should be there
    b.poisson = Poisson(mesh)
    b.poisson.bc = [AppliedVoltage(boundary) for boundary in mesh.boundaries]
    # All species are uncharged

    def v_D(ctx, eq):
        return 0., kwargs[eq.name][0]
    s = dict()
    for k in sorted(kwargs.keys()):
        species = AdvectionDiffusion(mesh, k, z=0, v_D=v_D)
        b.species.append(species)
        if bc_equation is not None:
            species.bc = bc_equation(species)
        s[k] = species
    b.setUp()
    params = {'T': 300., 'electrode0.voltage': 0, 'electrode1.voltage': 0, 'electrode0.workfunction': 0,
              'electrode1.workfunction': 0, 'epsilon_r': 3.}
    xinit = b.X.copy()
    for k, v in kwargs.items():
        xinit[s[k].idx] = kwargs[k][1](s[k].mesh.cells['center'] / L)
    conservation = []

    def update_conservation(t, out):
        for eq in b.species:
            store(out[eq.name + '.c'], rtol=1e-5, atol=1e-20)
        cons = [integrate_c(eq, out[eq.name + '.c']) for eq in b.species]
        for species in b.species:
            conservation.append((t,) + tuple(cons))
    update_conservation(0., b.output(0, xinit, 0. * xinit, params))
    c = context(b, x=xinit)
    for ts in c.transientsolve(params, times):
        pass
    for ts in c.timesteps():
        update_conservation(ts.time, ts.output())
        if make_plots:
            ylim = [0., np.amax(ts.x) * 1.1]
            p = mpl.forcontext(ts)
            p.allspecies(settings={})
            plt.ylim(ylim)
            plt.title('t=%e' % ts.time)
            plt.show()
    conservation = np.asarray(conservation)
    for i, species in enumerate(b.species):
        assert np.allclose(
            conservation[:, i + 1], conservation[0, i + 1], rtol=rtol, atol=atol)
    if make_plots:
        for i, species in enumerate(b.species):
            plt.plot(conservation[:, 0], conservation[:, i + 1], 'o-')
        plt.xscale('log')
        plt.title('Conservation of species')
        plt.xlabel('Time')
        plt.ylabel('Total amount of species')
        plt.show()
    if return_context:
        return c, conservation


test_species = dict(part1=(1., lambda u: np.abs(u - 0.5) < 0.1), part2=(10., lambda u: np.abs(u - 0.2) < 0.1),
                    part3=(1e2, lambda u: np.where(u > 0.999, 1e2, 0)))


class TestDiffusion(unittest.TestCase):
    _multiprocess_can_split_ = True

    def test_isolated(self):
        return run_diffusion(1., bc_equation=lambda eq: [], integrate_c=integrate_c_all, times=[
                             1e-9, 1e-7, 1e-5, 1e-3, 1e-1, 1e1], **test_species)

    def test_zero(self):
        return run_diffusion(1., bc_equation=lambda eq: [Zero('electrode0')], integrate_c=integrate_c_all, times=[
                             1e-9, 1e-7, 1e-5, 1e-3, 1e-1, 1e1], atol=1e100, **test_species)

    def test_periodic(self):
        return run_diffusion(1., bc_equation=lambda eq: [Equal(eq, 'electrode1')], integrate_c=integrate_c_all, times=[
                             1e-9, 1e-7, 1e-5, 1e-3, 1e-1, 1e1], **test_species)

    def test_periodic2(self):
        return run_diffusion(1., bc_equation=lambda eq: [Equal(eq, 'electrode0')], integrate_c=integrate_c_all, times=[
                             1e-9, 1e-7, 1e-5, 1e-3, 1e-1, 1e1], part1=(1., lambda x: np.exp(-4 * (x - 0.25)**2)), part2=(5., lambda x: np.cos(20 * x)**4))

    def test_periodic3(self):
        return run_diffusion(1., bc_equation=lambda eq: [Equal(eq, 'electrode1')], integrate_c=integrate_c_all, times=[
                             1e-9, 1e-7, 1e-5, 1e-3, 1e-1, 1e1], part1=(1., lambda x: np.exp(-4 * (x - 0.25)**2)), part2=(5., lambda x: np.cos(20 * x)**4))

    def test_mixed(self):
        return run_diffusion(1., bc_equation=lambda eq: [Zero('electrode1'), Equal(eq, 'electrode0')], integrate_c=integrate_c_all, times=[
                             1e-9, 1e-7, 1e-5, 1e-3, 1e-1, 1e1], part1=(1., lambda x: np.exp(-4 * (x - 0.25)**2)), part2=(5., lambda x: np.cos(20 * x)**4), atol=1e100)

    def runTest(self):
        pass



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = [
    'AdvectionDiffusion',
    'species_v_D_charged_from_params',
    'species_v_D_FrenkelPoole',
    'species_v_D_limited']

from oedes.ad import exp, where, sqrt
import numpy as np
from .transport import TransportedSpecies
from .frenkelpoole import FrenkelPooleMobility


class AdvectionDiffusion(TransportedSpecies):
    "General advection diffusion equation"

    def __init__(self, *args, **kwargs):
        v_D = kwargs.pop('v_D', None)
        super(AdvectionDiffusion, self).__init__(*args, **kwargs)
        self.v_D = v_D

    def evaluate_mobility(self, ctx, eq):
        ctx.varsOf(eq)['v_D'] = self.v_D(ctx, eq)


def species_v_D_charged_from_params(ctx, eq):
    """
    Simple transport model:
    - assumes diffusion described by Einstein's relation
    - mobility given as parameter (mu), or as a function f(T,c,E)
    """
    mu = ctx.param(eq, 'mu')
    if callable(mu):
        mu_cell = mu(
            T=ctx.varsOf(
                eq.thermal)['T'],
            c=ctx.varsOf(eq)['c'],
            E=ctx.varsOf(
                eq.poisson)['Ecellm'])
        mu = eq.mesh.faceaverage(mu_cell)
        ctx.varsOf(eq).update(mu_face=mu, mu_cell=mu_cell)
    else:
        ctx.varsOf(eq).update(mu_face=mu, mu_cell=mu)
    v = eq.v(mu, ctx.varsOf(eq.poisson)['E'])
    D = eq.D(mu, ctx.varsOf(eq.thermal)['Vt'])
    return v, D


def species_v_D_FrenkelPoole(ctx, eq):
    "Frenkel - Poole mobility model"
    FrenkelPooleMobility().mobility(None, ctx, eq)
    mu_face = ctx.varsOf(eq)['mu_face']
    mu_cell = ctx.varsOf(eq)['mu_cell']
    v = eq.v(mu_face, ctx.varsOf(eq.poisson)['E'])
    D = eq.mesh.faceaverage(eq.D(mu_cell, ctx.varsOf(eq.thermal)['Vt']))
    return v, D


def species_v_D_limited(ctx, eq):
    "Species with concentration limitation"
    v, D = species_v_D_charged_from_params(ctx, eq)
    c = ctx.varsOf(eq)['c']
    c = where(v > 0, c[eq.mesh.faces['+']], c[eq.mesh.faces['-']])
    v = v * (1. - c / ctx.param(eq, 'N0'))
    return v, D



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .weaklist import *
from .funcall import *
from .computation import *
from .builder import *
from .equation import *
from .evaluation import *



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# From https://www.nrel.gov/grid/solar-resource/spectra-am1.5.html

from oedes.optical.func import InterpolatedFunctionOfWavelength
import os
from six import BytesIO
import numpy as np
import zipfile

__all__ = ['AM0', 'AM1_5', 'AM1_5_global', 'AM1_5_direct']


def load():
    path = os.path.dirname(__file__)
    c = zipfile.ZipFile(os.path.join(path, 'astmg173.zip'))
    csv_data = c.read('ASTMG173.csv')
    data = np.genfromtxt(BytesIO(csv_data), delimiter=',', skip_header=2)
    data[:, 0] *= 1e-9
    data[:, 1:] *= 1e9
    AM0 = InterpolatedFunctionOfWavelength(data[:, 0], data[:, 1])
    AM1_5_global = InterpolatedFunctionOfWavelength(data[:, 0], data[:, 2])
    AM1_5_direct = InterpolatedFunctionOfWavelength(data[:, 0], data[:, 3])
    return AM0, AM1_5_global, AM1_5_direct


AM0, AM1_5_global, AM1_5_direct = load()
AM1_5 = AM1_5_global



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from . import ad
import numpy as np
import scipy.sparse
import warnings


class model(object):
    X = None
    transientvar = None

    def evaluate(self, time, x, xt, params, full_output=None, solver=None):
        raise NotImplemented()

    def residuals(self, time, x, xt, params, full_output=None, solver=None):
        return ad.sparsesum(len(self.X), self.evaluate(
            time, x, xt, params, full_output=full_output, solver=solver))

    def output(self, time, x, xt, params, solver=None):
        output = {}
        for _ in self.evaluate(time, x, xt, params,
                               full_output=output, solver=solver):
            pass
        return output

    def converged(self, residuals, x, dx, params, report):
        raise NotImplemented()

    def update(self, x, params):
        pass

    def scaling(self, params):
        return 1., 1.



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import base64
import json
import numpy as np
from contextlib import contextmanager
import os
import fnmatch
import subprocess
import tempfile
import time
import functools

array_hooks = []


@contextmanager
def printall():
    opt = np.get_printoptions()
    np.set_printoptions(threshold=np.inf)
    yield
    np.set_printoptions(**opt)


def array2dict(obj):
    assert isinstance(obj, np.ndarray)
    obj_data = np.ascontiguousarray(obj)
    return dict(__ndarray__=base64.b64encode(obj_data).decode('ascii'),
                dtype=str(obj.dtype),
                shape=",".join(map(str, obj.shape)))


def isarray(obj):
    return isinstance(obj, dict) and '__ndarray__' in obj


def dict2array(dct):
    assert isarray(dct)
    data = base64.b64decode(dct['__ndarray__'])
    shape = [int(s) for s in dct['shape'].split(',')]
    return np.frombuffer(data, np.dtype(dct['dtype'])).reshape(shape)


def nb_store_array(x, atol=0., rtol=1e-7, label=None):
    for hook in array_hooks:
        if hook(x, atol=atol, rtol=rtol, label=label):
            return
    x = np.asarray(x)
    if x.dtype == np.longdouble:
        x = np.asarray(x, dtype=np.double)
    elif x.dtype == np.longcomplex:
        x = np.asarray(x, dtype=np.complex)
    try:
        from IPython.display import publish_display_data
    except BaseException:
        return
    d = array2dict(x)
    u = dict2array(d)
    assert (x == u).all()
    data = {'application/json': dict(data=array2dict(x))}
    metadata = dict(format='oedes.testing.nb_store_array',
                    atol=atol, rtol=rtol, label=label)
    publish_display_data(data=data, metadata=metadata)


store = nb_store_array


class _store_result(object):
    def __init__(self, func, **kwargs):
        self.func = func
        self.store_kwargs = kwargs

    def __call__(self, *args, **kwargs):
        result = self.func(*args, **kwargs)
        nb_store_array(result, **self.store_kwargs)
        return result


def stored(**kwargs):
    return functools.partial(_store_result, **kwargs)


def stored_arrays(cells):
    for cell in cells:
        if cell.cell_type != 'code':
            continue
        for output in cell.outputs:
            try:
                json = output.data['application/json'].data
                meta = output.metadata
                if meta.format != 'oedes.testing.nb_store_array' or not isarray(
                        json):
                    continue
            except KeyError:
                continue
            except AttributeError:
                continue
            yield dict((k, v) for k, v in meta.items() if k != 'format'), dict2array(json)


def check_notebook_arrays(reference, tested):
    ro = list(stored_arrays(reference.cells))
    to = list(stored_arrays(tested.cells))
    if len(ro) != len(to):
        raise RuntimeError(
            'different number of arrays between reference and tested')
    for iarr, (ref, test) in enumerate(zip(ro, to)):
        rmeta, rarr = ref
        tmeta, tarr = test
        if not np.allclose(rarr, tarr, atol=rmeta['atol'], rtol=rmeta['rtol']):
            with printall():
                raise RuntimeError('\n'.join(
                    ["Difference in stored array %d" % iarr,
                     "Maximum error is %e" % np.amax(np.abs(rarr - tarr)),
                     "Maximum relative error is %e" % np.amax(
                         np.abs((rarr - tarr) / rarr)),
                     "Assumed rtol=%e atol=%e" % (
                         rmeta['rtol'], rmeta['atol']),
                     "Reference:",
                     str(rarr),
                     "Tested:",
                     str(tarr)
                     ]))
    return True


def read_notebook(fn):
    from nbformat.converter import convert
    from nbformat.reader import reads
    IPYTHON_VERSION = 4
    NBFORMAT_VERSION = 4

    with open(fn, 'r') as f:
        text = f.read()
        data = reads(text)
    return text, convert(data, NBFORMAT_VERSION)


def run_notebook(fn, timeout):
    _, reference = read_notebook(fn)
    temporary = tempfile.NamedTemporaryFile(suffix='.ipynb', delete=False)
    temporary.close()
    try:
        tempdir, tempname = os.path.split(temporary.name)
        for attempt in range(10):
            process = subprocess.Popen(['jupyter', 'nbconvert', '--ExecutePreprocessor.timeout=%d' % timeout, '--ExecutePreprocessor.kernel_name=python',
                                        '--to', 'notebook', '--execute', fn, '--output', tempname, '--output-dir', tempdir], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()
            if stdout is not None:
                stdout = stdout.decode()
            if stderr is not None:
                stderr = stderr.decode()
                if stderr.find('ZMQError: Address already in use') >= 0:
                    time.sleep(np.random.rand() + 0.1)
                    continue
            if process.returncode != 0:
                raise RuntimeError("\n".join(['Unable to execute notebook (returned exitcode %d)' % process.returncode,
                                              'Output:',
                                              stdout,
                                              'Errors:',
                                              stderr]))
            text, tested = read_notebook(temporary.name)
            return reference, tested, text
        raise RuntimeError(
            'workaround: maximum number of %d attempts reached' % attempt)
    finally:
        os.remove(temporary.name)

# def run_notebook(fn, **kwargs):
#    import nbformat
#    from nbconvert.preprocessors import ExecutePreprocessor
#    import zmq
#    with open(fn,'r') as f:
#        ref=nbformat.read(f,as_version=4)
#    for attempt in range(10):
#        try:
#            ep=ExecutePreprocessor(kernel_name='python',**kwargs)
#            result,res=ep.preprocess(ref,{})
#            break
#        except zmq.ZMQError:
#            continue
#    return ref,result,nbformat.writes(result)


def find_notebooks(rootdir='.'):
    for root, dirnames, filenames in os.walk(rootdir):
        root = os.path.relpath(root, rootdir)
        if any([x.startswith('.') for x in os.path.split(root)]):
            continue
        for filename in fnmatch.filter(filenames, '*.ipynb'):
            f = os.path.join(root, filename)
            if os.path.islink(f):
                continue
            yield os.path.relpath(f, rootdir)


from oedes.fvm import mesh1d
from oedes.models import electronic_device, BaseModel


def unipolar(L, polarity='p'):
    mesh = mesh1d(L)
    model = BaseModel()
    electronic_device(model, mesh, polarity)
    prefix = model.species[0].name
    params = {
        'T': 300,
        'epsilon_r': 3.,
        'electrode0.voltage': 0,
        'electrode1.voltage': 0,
        'electrode0.workfunction': 0,
        'electrode1.workfunction': 0,
        prefix + '.energy': 0,
        prefix + '.N0': 1e27,
        prefix + '.mu': 1e-9
    }
    return model, params


from oedes.fvm import multilayer1d
from oedes.models import CompositeModel, Equal, FermiLevelEqual, FermiLevelEqualElectrode, AppliedVoltage


def bilayer(L0, L1, polarity='p'):
    ml = multilayer1d([('layer0', mesh1d(L0, boundary_names=['electrode0', 'internal'])),
                       ('layer1', mesh1d(L1, boundary_names=['internal', 'electrode1']))])
    model = CompositeModel()
    params = {
        'layer0.electrode0.voltage': 0,
        'layer1.electrode1.voltage': 0,
        'layer0.electrode0.workfunction': 0,
        'layer1.electrode1.workfunction': 0
    }
    if 'n' in polarity and 'p' in polarity:
        params.update({
            'layer0.npi': 0,
            'layer1.npi': 0
        })
    for i, name in enumerate(ml.domains):
        sub = BaseModel()
        sub.name = name
        model.sub.append(sub)
        electronic_device(sub, ml.domains[name], polarity)
        ename = name.replace('layer', 'electrode')
        sub.poisson.bc = [AppliedVoltage(ename)]
        params.update({name + '.T': 300,
                       name + '.epsilon_r': 3.})
        for s in sub.species:
            s.bc = [FermiLevelEqualElectrode(ename)]
            prefix = name + '.' + s.prefix
            params.update({
                prefix + '.energy': 0,
                prefix + '.N0': 1e27,
                prefix + '.mu': 1e-9
            })
    l0, l1 = model.sub
    for i, s in enumerate(l1.species):
        s.bc.append(FermiLevelEqual(l0.species[i], 'internal'))
    l1.poisson.bc.append(Equal(l0.poisson, 'internal'))
    return model, params



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# This tests if Poisson's equation is solved correctly in most simple cases

from oedes import *
from oedes.fvm import mesh1d
from oedes.models import BaseModel, Poisson, AppliedVoltage
L = 100e-9
v0 = 1.
v1 = -1.
mesh = mesh1d(L)


def run_poisson(bc):
    b = BaseModel()
    b.poisson = Poisson(mesh)
    b.poisson.bc = bc
    params = {'T': 300., 'electrode0.voltage': v0, 'electrode1.voltage': v1,
              'electrode0.workfunction': 0., 'electrode1.workfunction': 0., 'epsilon_r': 3.}
    b.setUp()
    solution = solve(b, b.X, params)
    out = b.output(0., solution, 0. * b.X, params)
    return out


def test_poisson_DirichletDirichlet():
    out = run_poisson([AppliedVoltage('electrode0'),
                       AppliedVoltage('electrode1')])
    assert np.allclose(out['E'], (v0 - v1) / L)
    assert np.allclose(out['potential'], v0 + (v1 - v0)
                       * mesh.cells['center'] / L)


def test_poisson_DirichletOpen():
    out = run_poisson([AppliedVoltage('electrode0')])
    assert np.allclose(out['E'], 0.)
    assert np.allclose(out['potential'], v0)


def test_poisson_OpenDirichlet():
    # # Open-Dirichlet
    out = run_poisson([AppliedVoltage('electrode1')])
    assert np.allclose(out['E'], 0.)
    assert np.allclose(out['potential'], v1)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .charged import ChargedSpecies
from oedes.fvm import DummyConvergenceTest

__all__ = ['FixedCharge']


class FixedCharge(ChargedSpecies):
    "Fixed charge, typically corresponding to ionized dopants"

    def __init__(self, poisson, mesh=None, name=None, z=1, density=None):
        super(FixedCharge, self).__init__(mesh, name, z, poisson=poisson)
        if density is not None:
            self.density = density

    def density(self, x, ctx, eq):
        return ctx.param(eq, 'Nd') - ctx.param(eq, 'Na')

    def buildDiscreteEq(self, ctx, eq):
        super(FixedCharge, self).buildDiscreteEq(ctx, eq)
        eq.convergenceTest = DummyConvergenceTest()

    def load_concentration(self, ctx, eq):
        c = self.density(eq.mesh, ctx, eq)
        ctx.varsOf(eq).update(c=c, j=None)
        self.output_concentration(ctx, eq)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['BuilderData', 'FVMBuilder', 'discretize']

import numpy as np
from oedes.utils import BuilderData
from collections import defaultdict
from .cell import FVMBoundaryEquation, FVMConservationEquation
from .poisson import FVMPoissonEquation
from .transport import FVMTransportEquation, FVMTransportChargedEquation
from .discrete import GeneralDiscreteEquation, DiscreteEquation, DummyConvergenceTest, DelegateConvergenceTest
from .evaluator import FVMEvaluator
from oedes.utils import ArgStack, EvaluationContext, as_computation
from .mesh import mesh


def _EquationBasedOrdering(equations):
    i = 0
    for eq in equations:
        n = eq.ndof()
        eq.idx = np.arange(i, i + n)
        i += n
    return i


def _CellBasedOrdering(equations):
    count = defaultdict(int)
    for eq in equations:
        if isinstance(eq, FVMConservationEquation):
            count[id(eq.mesh)] += 1
    j = dict()
    i = 0
    for eq in equations:
        n = eq.ndof()
        if not isinstance(eq, FVMConservationEquation):
            eq.idx = np.arange(i, i + n)
            i += n
        else:
            k = id(eq.mesh)
            if k not in j:
                j[k] = i
                i += eq.mesh.ncells * count[id(eq.mesh)]
            assert eq.ndof() == eq.mesh.ncells
            eq.idx = j[k] + np.arange(eq.mesh.ncells) * count[k]
            j[k] += 1
    return i


class FVMBuilder(BuilderData):
    def __init__(self, ordering='equation', mesh_mapper=None):
        super(FVMBuilder, self).__init__()
        if ordering == 'cell':
            indexer = _CellBasedOrdering
        elif ordering == 'equation':
            indexer = _EquationBasedOrdering
        else:
            raise ValueError('unknown equation order - %r' % ordering)
        self.indexer = indexer
        self.mesh_mapper = mesh_mapper

    def getMesh(self, key):
        if isinstance(key, mesh):
            return key
        return self.mesh_mapper(key)

    def setUp(self, context_type):
        self.ndof = self.indexer(self.equations)
        self.initEquations()
        self.evaluator = FVMEvaluator(self, context_type)
        self.finalize(self.evaluator)

    def newPoissonEquation(self, *args, **kwargs):
        return FVMPoissonEquation(*args, **kwargs)

    def newTransportEquation(self, *args, **kwargs):
        return FVMTransportEquation(*args, **kwargs)

    def newTransportChargedEquation(self, *args, **kwargs):
        return FVMTransportChargedEquation(*args, **kwargs)

    def newDirichletBC(self, *args, **kwargs):
        return FVMBoundaryEquation(*args, **kwargs)

    def newGeneralDiscreteEquation(self, *args, **kwargs):
        return GeneralDiscreteEquation(*args, **kwargs)

    def add(self, owner, eq):
        assert isinstance(eq, DiscreteEquation)
        assert owner is not None
        if isinstance(eq, FVMBoundaryEquation):
            assert eq.convergenceTest is not None
        super(FVMBuilder, self).add(owner, eq)

    def newDelegateConvergenceTest(self, obj):
        if obj is None:
            return DummyConvergenceTest()
        else:
            return DelegateConvergenceTest(obj)


def _meshdict(arg):
    if isinstance(arg, mesh):
        d = {None: arg}
    else:
        d = arg
    return lambda x: d[x]


def discretize(obj, mesh=None, **kwargs):
    obj = as_computation(obj)
    builder = FVMBuilder(mesh_mapper=_meshdict(mesh), **kwargs)
    obj.discretize(builder)
    builder.setUp(EvaluationContext)
    for _, eq in obj.all_equations(ArgStack()):
        eq.idx = builder.get(eq).idx  # TODO:
    return builder.evaluator



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
from .cell import FVMBoundaryEquation, FVMConservationEquation
import scipy.sparse.csgraph
from oedes.ad import sparsesum
from oedes.model import model


class FVMEvalContext(object):
    def __init__(self, evaluator, target):
        self.target = target
        self.evaluator = evaluator
        self.bc_conservation = []

    def finalize(self):
        return self.evaluator.finalize(self.target, self.bc_conservation)


class KillVars(object):
    def __init__(self, obj):
        self.obj = obj

    def __call__(self, ctx):
        ctx.killVars(self.obj)


class FVMEvaluator(model):
    def __init__(self, builder, context_type):
        self.equations = builder.equations
        self.ndof = builder.ndof

        self._init_boundary()

        X = np.zeros(builder.ndof)
        self.X = X

        transientvar = X.copy()
        for eq in self.equations:
            transientvar[eq.idx] = eq.transientvar
        self.transientvar = transientvar
        self.plan = tuple(builder.fungraph.plan(kill_func=KillVars))
        self.context_type = context_type

    def _all_bc(self):
        for eq in self.equations:
            if isinstance(eq, FVMBoundaryEquation):
                yield eq

    def _all_conservation(self):
        for eq in self.equations:
            if isinstance(eq, FVMConservationEquation):
                yield eq

    def _init_boundary(self):
        i = [np.zeros(0, dtype=np.int)]
        j = [np.zeros(0, dtype=np.int)]
        for bc in self._all_bc():
            valid = bc.conservation_to_dof >= 0
            i.append(bc._getDof()[valid])
            j.append(bc.conservation_to_dof[valid])
            #eq.bc_free_dof = np.arange(len(eq.mesh.boundary.idx))[fdof]
        i = np.hstack(i)
        j = np.hstack(j)
        # optimize: should only create vector for boundary items
        g = scipy.sparse.csr_matrix(
            (np.ones_like(i), (i, j)), shape=(self.ndof,) * 2)
        nlabels, labels = scipy.sparse.csgraph.connected_components(
            g, directed=False, return_labels=True)
        self.bc_label_volume = sparsesum(nlabels, ((labels[eq.idx], eq.mesh.cells[
                                         'volume']) for eq in self.equations if isinstance(eq, FVMConservationEquation)))
        self.bc_labels = labels
        for eq in self._all_conservation():
            eq.boundary_labels = self.bc_labels[eq.idx[eq.mesh.boundary.idx]]

    def createContext(self, target):
        return FVMEvalContext(self, target)

    def finalize(self, target, bc_conservation):
        bc_label_conservation = sparsesum(
            len(self.bc_labels), bc_conservation)
        for eq in self.equations:
            if isinstance(eq, FVMConservationEquation):
                bc_free_dof = np.arange(len(eq.mesh.boundary.idx))[
                    eq.bc_dof_is_free]
                i = eq.idx[eq.mesh.boundary.idx[bc_free_dof]]
                j = self.bc_labels[i]
                yield i, bc_label_conservation[j] * (1. / self.bc_label_volume[j])

    def scaling(self, params):
        xscaling = np.ones_like(self.X)
        fscaling = np.ones_like(self.X)
        for eq in self.equations:
            eq.scaling(xscaling, fscaling)
        return xscaling, fscaling

    def update(self, x, params):
        for eq in self.equations:
            eq.update(x, params)

    def evaluate(self, time, x, xt, params, full_output=None, solver=None):
        ectx = self.createContext(None)
        ctx = self.context_type(
            params,
            full_output,
            gvars=dict(
                time=time,
                x=x,
                xt=xt,
                params=params,
                full_output=full_output,
                solver=solver,
                bc_conservation=ectx.bc_conservation))

        for p in self.plan:
            itr = p(ctx)
            if itr is not None:
                for ix, v in itr:
                    yield ix, v
        for ix, v in ectx.finalize():
            yield ix, v
        # for ix,v in self.generate_bc_conservation(ctx):
        #    yield ix,v

    def converged(self, residuals, x, dx, params, report, equations=None):
        if equations is None:
            equations = self.equations
        c = all([eq.testConverged(residuals, x, dx, report)
                 for eq in equations])
        return c



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.ad import getitem, sparsesum, dot
from oedes.utils import EquationWithMesh, Equation
import itertools
import weakref

__all__ = ['ConservationEquation', 'BoundaryEquation']


class ConservationEquation(EquationWithMesh):
    def __init__(self, mesh=None, name=None):
        super(ConservationEquation, self).__init__(mesh, name)
        self.defaultBCConvergenceTest = None
        self.bc = []

    def load(self, ctx, eq):
        super(ConservationEquation, self).load(ctx, eq)
        vars = ctx.varsOf(eq)
        vars['boundary_sources'] = []
        vars['boundary_FdS'] = []

    def _evaluate_bc(self, ctx, eq, FdS_boundary,
                     celltransient_boundary=0., cellsource_boundary=0.):
        conservation = -FdS_boundary + eq.mesh.cells['volume'][
            eq.mesh.boundary.idx] * (celltransient_boundary - cellsource_boundary)
        ctx.gvars['bc_conservation'].append((eq.boundary_labels, conservation))

    def residuals(self, ctx, eq, flux, source=None, transient=0.):
        variables = ctx.varsOf(eq)
        yield eq.residuals(eq.mesh.internal, flux, cellsource=source, celltransient=transient)
        n = len(eq.mesh.boundary.cells)
        bc_FdS = sparsesum(n, variables['boundary_FdS'])
        bc_source = sparsesum(n, variables['boundary_sources'])
        if flux is not None:
            bc_FdS = dot(eq.mesh.boundary.fluxsum, flux) + bc_FdS
        if source is not None:
            bc_source = getitem(source, eq.mesh.boundary.idx) + bc_source
        bc_transient = getitem(transient, eq.mesh.boundary.idx)
        variables['total_boundary_FdS'] = bc_FdS
        variables['total_boundary_sources'] = bc_source
        variables['total_boundary_transient'] = bc_transient
        self._evaluate_bc(
            ctx,
            eq,
            bc_FdS,
            cellsource_boundary=bc_source,
            celltransient_boundary=bc_transient)

    def identity(self, ctx, eq):
        yield eq.identity(ctx.varsOf(eq)['x'])

    def all_equations(self, args):
        def bcs():
            for bc in self.bc:
                yield bc
        for eq in super(ConservationEquation, self).all_equations(args):
            yield eq
        for bc in self.bc:
            assert bc._owner_eq is None
            if bc._owner_eq_weak is None:
                bc._owner_eq_weak = weakref.ref(self)
            else:
                assert bc._owner_eq_weak() is self
            for eq in bc.all_equations(args):
                yield eq


class BoundaryEquation(Equation):
    def __init__(self, name, owner=None):
        super(BoundaryEquation, self).__init__(name)
        self._owner_eq_weak = None
        self._owner_eq = owner

    @property
    def owner_eq(self):
        if self._owner_eq is not None:
            assert self._owner_eq_weak is None, 'cannot assign both references to owner'
            return self._owner_eq
        assert self._owner_eq_weak is not None, 'no reference to owner'
        return self._owner_eq_weak()

    def newDiscreteEq(self, builder):
        return builder.newDirichletBC(builder.get(self.owner_eq), self.name)

    def evaluate(self, ctx, eq):
        raise NotImplementedError()

    def initDiscreteEq(self, builder, obj):
        super(BoundaryEquation, self).initDiscreteEq(builder, obj)
        obj.convergenceTest = builder.newDelegateConvergenceTest(self.owner_eq)
        obj.owner_eq = builder.get(self.owner_eq)
        obj.evaluate.depends(obj.owner_eq.load)
        obj.owner_eq.alldone.depends(obj.evaluate)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['CompositeModel', 'With', 'Coupled']

from oedes import model
from oedes.fvm import discretize
from oedes.utils import With, Coupled
from .solver import SolverObject


class ModelAdapter(model):
    def __init__(self):
        super(ModelAdapter, self).__init__()
        self.discrete_model = None
        self.ordering = 'equation'

    def scaling(self, params):
        return self.discrete_model.scaling(params)

    def update(self, x, params):
        return self.discrete_model.update(x, params)

    def converged(self, residuals, x, dx, params, report):
        return self.discrete_model.converged(residuals, x, dx, params, report)

    @property
    def X(self):
        return self.discrete_model.X

    @property
    def transientvar(self):
        return self.discrete_model.transientvar

    def evaluate(self, time, x, xt, params, full_output=None, solver=None):
        if solver is None:
            solver = SolverObject()
        return self.discrete_model.evaluate(
            time, x, xt, params, full_output=full_output, solver=solver)

    def setUp(self, *args, **kwargs):
        assert self.discrete_model is None, 'already set-up'
        self.discrete_model = discretize(
            self, ordering=self.ordering, *args, **kwargs)


class CompositeModel(ModelAdapter, Coupled):
    def __init__(self, parts=None):
        super(CompositeModel, self).__init__()
        self.sub = []
        self.other = []
        if parts is not None:
            self.other.extend(parts)

    @property
    def parts(self):
        for sub in self.sub:
            yield With(sub, name=sub.name)
        for other in self.other:
            yield other



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import collections
from .computation import Computation

__all__ = ['OutputMeta', 'Units', 'FreedVars', 'EvaluationContext', 'meta_key']

OutputMeta = collections.namedtuple('OutputMeta', ['mesh', 'face', 'unit'])


class Units:
    V = 'V'
    potential = V
    charge_density = 'C/m^3'
    electric_field = 'V/m'
    current_density = 'A/m^2'
    concentration = '1/m^3'
    dconcentration_dt = '1/(m^3 s)'
    flux = '1/(m^3 s)'
    eV = 'eV'
    meter = 'm'

    @classmethod
    def check(cls, unit):
        assert unit in [
            None,
            cls.V,
            cls.charge_density,
            cls.electric_field,
            cls.current_density,
            cls.concentration,
            cls.flux,
            cls.eV,
            cls.dconcentration_dt]


meta_key = '.meta'


class FreedVars(object):
    pass


class EvaluationContext(object):
    UPPER = '..'

    units = Units

    def __init__(self, params, outputs, vardict=None, gvars=None):
        if vardict is None:
            vardict = dict()
        assert gvars is not None
        self._vardict = vardict
        self.params = params
        self.outputs = outputs
        self.gvars = gvars

        # TODO:
        if self.outputs is not None:
            if meta_key not in self.outputs:
                self.outputs[meta_key] = dict()
            self.outputs_meta = self.outputs[meta_key]

    def newVars(self, obj, varobj=None):
        if varobj is None:
            varobj = dict()
        assert not isinstance(obj, Computation)
        k = id(obj)
        assert k not in self._vardict
        self._vardict[k] = varobj
        return varobj

    def varsOf(self, obj):
        assert not isinstance(obj, Computation)
        return self._vardict[id(obj)]

    def killVars(self, obj):
        assert not isinstance(obj, Computation)
        self._vardict[id(obj)] = FreedVars()

    @property
    def wants_output(self):
        return self.outputs is not None

    def _path(self, eq, *args):
        if eq is not None and eq.prefix:
            path = list(eq.prefix.split('.'))
        else:
            path = []
        for a in args:
            if a is self.UPPER:
                path = path[:-1]
            else:
                path.append(a)
        return '.'.join(path)

    def _absdict(self, obj):
        return obj

    def param(self, *args):
        p = self._path(*args)
        return self._absdict(self.params)[p]

    def output(self, args, value, mesh=None, face=False, unit=None, add=False):
        if self.outputs is None:
            return
        p = self._path(*args)
        d = self._absdict(self.outputs)
        meta = self._absdict(self.outputs_meta)
        if add and p in d:
            d[p] = d[p] + value
        else:
            assert p not in d, 'duplicate assignment to %r' % p
            d[p] = value
        Units.check(unit)
        meta_info = OutputMeta(mesh=mesh, face=face, unit=unit)
        if add and p in meta:
            assert meta[p] == meta_info
        else:
            meta[p] = meta_info

    def common_param(self, eqs, *args):
        p = self._path(eqs[0], self.UPPER, *args)
        assert all([self._path(e, self.UPPER, *args) == p for e in eqs[1:]])
        return self._absdict(self.params)[p]

    def _meshOutput(self, args, value, mesh=None, unit=None, face=None):
        if mesh is None:
            mesh = args[0].mesh
        return self.output(args, value, mesh=mesh, face=face, unit=unit)

    def outputFace(self, args, value, **kwargs):
        return self._meshOutput(args, value, face=True, **kwargs)

    def outputCell(self, args, value, **kwargs):
        return self._meshOutput(args, value, face=False, **kwargs)

    def __getattr__(self, name):
        if name in self.gvars:
            return self.gvars[name]
        else:
            return super(EvaluationContext, self).__getattr__(self, name)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes import ad
import numpy as np
from . import gdos
from .physics import spacing_to_N0 as atoN0
from .physics import N0_to_spacing as N0toa

# In all functions, recommended 3<=nsigma<=6

g1_max_c = 0.1
g2_max_En = 2.
g2_max_En2 = g2_max_En**2
g3_max_c = 0.5


def g1(nsigma, c):
    """
    EGDM g1
    nsigma=sigma/kT: must be nsigma>=1.
    c must be 0<=c<=1, recommended c<0.1
    """
    delta = 2.0 * (ad.log(nsigma**2 - nsigma) -
                   ad.log(ad.log(4))) / (nsigma**2)
    return ad.exp(0.5 * (nsigma**2 - nsigma) * (2 * c)**delta)


def g2_En2(nsigma, En2):
    """
    EGDM g2(En2)
    nsigma=sigma/kT, recommended 3<=nsigma<=6
    En2=(E*a/sigma)**2, recommended En2<4.
    """
    return ad.exp(0.44 * (nsigma**(3. / 2.) - 2.2)
                  * (ad.sqrt(1 + 0.8 * En2) - 1))


def g2(nsigma, En):
    """EGDM g2(En), nsigma=sigma/kT, En=E*a/sigma"""
    return g2_En2(nsigma, En**2)


def mu0t(nsigma, mu0):
    """EGDM mu_0(T) - temperature dependent mobility prefactor in
    function of temperature independent prefactor"""
    return mu0 * 1.8e-9 * ad.exp(-0.42 * nsigma**2)


def mu0t_inverse(nsigma, mu0t):
    return mu0t / (1.8e-9 * ad.exp(-0.42 * nsigma**2))


def nsigma(sigma, Vt):
    """Return nsigma=sigma/kT, sigma is in eV, thermal voltage Vt=kT/q is in V"""
    return sigma / Vt


def nsigma_inverse(nsigma, Vt):
    return nsigma * Vt


def g3(nsigma, c, impl=gdos.defaultImpl, b=None):
    return gdos.diffusion_enhancement(nsigma * np.sqrt(2.), c, impl, b=b)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

class paramvector_params():
    def __init__(self, base, values, ix):
        self.base = base
        self.values = values
        self.ix = ix

    def __getitem__(self, u):
        if u in self.ix:
            return self.values[self.ix[u]]
        else:
            return self.base[u]


class paramvector():
    def __init__(self, names):
        assert len(set(names)) == len(names), 'do not duplicate'
        self.ordering = tuple(names)
        self.ix = dict((name, i) for (i, name) in enumerate(self.ordering))

    def values(self, params):
        return [params[name] for i, name in enumerate(self.ordering)]

    def asdict(self, base, values):
        return paramvector_params(base, values, self.ix)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import itertools

__all__ = ['ArgStack', 'Computation', 'With', 'Coupled', 'as_computation']


class ArgStack(object):
    stack_keys = [('prefix', 'name')]

    def __init__(self, **kwargs):
        for field, _ in self.stack_keys:
            if field not in kwargs:
                kwargs[field] = []
        self.kwargs = kwargs
        self._prefix = '.'.join(p for p in self.kwargs['prefix'])

    def push(self, **kwargs):
        for field, _ in self.stack_keys:
            assert field not in kwargs
        args = dict(self.kwargs)
        args.update(kwargs)
        for field, key in self.stack_keys:
            if key in kwargs and kwargs[key] is not None:
                args[field] = self.kwargs[field] + [kwargs[key]]
        return ArgStack(**args)

    @property
    def prefix(self):
        return self._prefix

    def __getattr__(self, name):
        return self.kwargs[name]


class Computation(object):
    def all_equations(self, args):
        yield (args, self)

    def build(self, builder):
        raise NotImplementedError()

    def findeq(self, name):
        return dict((a.prefix, eq)
                    for a, eq in self.all_equations(ArgStack()))[name]

    def findeqs(self, test, return_names=False):
        for a, eq in self.all_equations(ArgStack()):
            if test(eq):
                if return_names:
                    yield a.prefix, eq
                else:
                    yield eq


def as_computation(obj):
    "Returns object as calculation"
    if isinstance(obj, Computation):
        return obj
    return Coupled(obj)


class With(Computation):
    "Assings arguments (name etc.) to calculations"

    def __init__(self, eq, **kwargs):
        self.eq = as_computation(eq)
        self.kwargs = kwargs

    def all_equations(self, args):
        defaults = args.push(**self.kwargs)
        return self.eq.all_equations(defaults)


class Coupled(Computation):
    "Coupled calculations"

    def __init__(self, parts=None):
        if parts is not None:
            self.parts = tuple(map(as_computation, parts))

    def all_equations(self, args):
        return itertools.chain(*(p.all_equations(args) for p in self.parts))

    def discretize(self, builder):
        for args, eq in self.all_equations(ArgStack()):
            eq.build(builder, args)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = [
    'BaseModel',
    'std',
    'electronic_device',
    'add_transport',
    'add_ions',
    'electrolyte',
    'holeonly',
    'electrononly',
    'bulk_heterojunction',
    'bulk_heterojunction_params']

from .domain import Domain
from .equations import AdvectionDiffusion, WithDOS


class BaseModel(Domain):
    """
    Basic device model:
    Supports arbitrary number of species (transported or not) coupled to Poisson's equation, with any form of
    boundary conditions and drift and diffusion velocities
    """

    def __init__(self, *args, **kwargs):
        super(BaseModel, self).__init__(*args, **kwargs)
        self.species_v_D = dict()
        self.species_dos = dict()

    @property
    def parts(self):
        for eq in self.species:
            name = eq.name
            if isinstance(eq, WithDOS):
                if name in self.species_dos:
                    eq.dos = self._check(eq.dos, self.species_dos[name])
            else:
                assert name not in self.species_dos
            if isinstance(eq, AdvectionDiffusion) in self.species_v_D:
                eq.v_D = self._check(eq.v_D, self.species_v_D[name])
            else:
                assert name not in self.species_v_D
        return super(BaseModel, self).parts

    @property
    def equations(self):
        return self.parts


from . import std
from .std import *



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import scipy.interpolate
import numpy as np

__all__ = [
    'FunctionOfWavelength',
    'ConstFunctionOfWavelength',
    'InterpolatedFunctionOfWavelength',
    'wavelength_range',
    'makeComplexFunctionOfWavelength',
    'Material']


class FunctionOfWavelength(object):
    "This defines __call__(wavelength) and limits, with wavelength in m"
    pass


class InterpolatedFunctionOfWavelength(FunctionOfWavelength):
    def __init__(self, x, y):
        if np.any(np.diff(x) <= 0):
            raise ValueError('x values not sorted : {x}'.format(x=x))
        self.interp = scipy.interpolate.interp1d(
            x, y, kind='linear', bounds_error='True', assume_sorted=True)
        self.xdata = x
        self.ydata = y
        self.limits = (np.amin(x), np.amax(x))

    def __call__(self, wavelength):
        return self.interp(wavelength)


class ConstFunctionOfWavelength(FunctionOfWavelength):
    def __init__(self, value, limits=(-np.inf, np.inf)):
        self.value = value
        self.limits = np.asarray(limits)

    def __call__(self, wavelength):
        return self.value*np.ones_like(wavelength)


def wavelength_range(*functions):
    return tuple([max(u.limits[0] for u in functions),
                  min(u.limits[1] for u in functions)])


class makeComplexFunctionOfWavelength(object):
    def __init__(self, real_part, imag_part):
        self.limits = wavelength_range(real_part, imag_part)
        self.real_part = real_part
        self.imag_part = imag_part

    def __call__(self, wavelength):
        return self.real_part(wavelength) + self.imag_part(wavelength)*1.J


class Material(object):
    def __init__(self, refractive_index, name=None):
        self.refractive_index = refractive_index
        self.name = name



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
from .ad import *
from .param import paramvector
from sparsegrad.impl import scipy
if hasattr(scipy.sparse, 'linalg'):
    linalg = scipy.sparse.linalg
else:
    from scipy.sparse import linalg
import numpy as np
import warnings
import logging
from collections import defaultdict
from . import logs


class SolverError(RuntimeError):

    def __init__(self, *args):
        RuntimeError.__init__(self, *args)


class SolverObject(object):
    def __init__(self):
        pass

    @property
    def poissonOnly(self):
        return False


# In[41]:

# <api>


def _convert(A, b):
    if A.dtype == np.longdouble:
        A = scipy.sparse.csr_matrix(A, dtype=np.double)
    elif A.dtype == np.longcomplex:
        A = scipy.sparse.csr_matrix(A, dtype=np.complex)
    if b.dtype == np.longdouble:
        b = np.asarray(b, dtype=np.double)
    elif b.dtype == np.longcomplex:
        b = np.asarray(b, dtype=np.complex)
    return A, b


def spsolve_scipy_(A, b):
    A, b = _convert(A, b)
    with warnings.catch_warnings():
        warnings.simplefilter(
            "error", category=linalg.MatrixRankWarning)
        logs.linear.info('Solving linear equation using scipy sparse solver')
        logs.linear.info(
            'System matrix has shape {shape} with {nnz} nonzero entries of dtype {dtype}'.format(
                shape=A.shape, nnz=A.nnz, dtype=A.dtype))
        try:
            return linalg.spsolve(A, b, use_umfpack=False)
        except BaseException:
            raise SolverError('spsolve failed')


def spsolve_scipy(A, b):
    return spsolve_scipy_(A, b)


def _explain(report):
    columns = ['atol', 'rtol', 'maxabs', 'maxrel', 'maxabs_nc', 'maxrel_nc']
    header = '%-26s C?  atol    rtol   maxabs  maxrel nc_maxabs nc_maxrel' % 'equation'

    def makerow(k):
        d = report[k]
        if len(k) > 26:
            k = '>' + k[-25:]
        items = [k, ] + [d[u] for u in columns]
        numbers = tuple('{:^7}'.format('%1.1e' % d[u]) for u in columns)
        return '%-26s %.1s %s' % (k, d['converged'], ' '.join(numbers))
    return '\n'.join([header] + [makerow(k) for k in sorted(report.keys())])


def _matrixsolve(spsolve, A, b, scaling):
    xscaling, fscaling = scaling
    Mr = scipy.sparse.diags(
        [xscaling],
        offsets=[0],
        format='csr',
        dtype=xscaling.dtype)
    Ml = scipy.sparse.diags(
        [fscaling],
        offsets=[0],
        format='csr',
        dtype=fscaling.dtype)
    x_ = spsolve(Ml.dot(A.dot(Mr)), Ml.dot(b))
    return Mr.dot(x_)


def solve(model, x0, params, tconst=0., tshift=0., time=0.,
          maxiter=10, spsolve=spsolve_scipy, solver=None, niter=None):
    """This solves system F(x)=0 with x0 as initial guess

    Transient solution is possible by assuming xt=x*tshift+tconst.
    Both tshift and tconst should be zero for stationary.

    If should does not specify :convergence_test: (F,x,dx,params)

    Returns x,xt.
    """

    if solver is None:
        solver = SolverObject()

    F = model.residuals
    xscaling, fscaling = model.scaling(params)
    convergence_test = model.converged
    update = model.update

    logger = logs.nonlinear

    logs.timestepping.info('Solving %s' % solver)

    # _ denotes scaled variables
    x_ = x0 / xscaling
    dx_ = None
    itno = 0
    report = None
    timing_logger = logs.nonlinear.getChild('timing')
    while True:
        itno = itno + 1
        if not np.alltrue(np.isfinite(x_)):
            raise SolverError('solution diverged')
        if itno > maxiter:
            nclist = [k for k, v in report.items() if not v['converged']]
            logger.info(
                'Did not converge in %d iterations: %s' %
                (maxiter, nclist))
            raise SolverError('did not converge in %d iterations' % maxiter)
        timer = logs.Timer()
        adx = forward.seed_sparse_gradient(x_) * xscaling
        f_unscaled = F(time, adx, adx * tshift + tconst, params, solver=solver)
        f = f_unscaled * fscaling
        logger.info("Nonlinear iteration %d, residual norm |F|=%e (unscaled %e)" %
                    (itno, np.linalg.norm(f.value), np.linalg.norm(f_unscaled.value)))
        if dx_ is not None:
            report = dict()
            if convergence_test(f_unscaled.value, adx.value,
                                dx_ * xscaling, params, report=report):
                break
        timing_logger.info(
            'Calculation of Jacobian matrix and F took %s' %
            timer)
        timer = logs.Timer()
        dx_ = spsolve(f.gradient.tocsr(), -f.value)
        timing_logger.info('Linear solve took %s' % timer)
        if not np.alltrue(np.isfinite(dx_)):
            msg = 'linear solver failed (NaN)'
            logger.info(msg)
            raise SolverError(msg)
        x_ = x_ + dx_
        x = x_ * xscaling
        update(x, params)
        x_ = x / xscaling
        if niter is not None and itno >= niter:
            break
    x = x_ * xscaling
    #logger.info('converged in %d iterations' % itno)
    return x


def bdf1adapt_(model, x, params, t, t1, dt, mindt=1e-15, xt=None, dtrlim=(0.5, 2.), relfail=10., reltol=1e-2, abstol=1e-15, dtrfailsolve=0.1,
               dtrfaillte=0.1, weight=None, skip=2, use_predictor=False, final_matchstep=True, solve=solve, maxsteps=1000, maxdt=np.inf, solver=None, **kwargs):
    """
    Adaptive transient solver based on BDF1. Solve from time :t0: to at least :t1:, with initial timestep :dt:.
    :xt: is xt at t, can be None.
    :mindt: is minimum dt, below which calculation will be aborted.

    The goal is to keep norm of solution LTE (local truncation error) satisfying:
    norm(lte) <= reltol*norm(solution)+abstol

    Variables are taken into account scaled by :weight:.

    If norm(lte) > reject*(reltol*norm(solution)+abstol), the timestep is rejected and new dt tried faillte.

    The timestep in after successful step is in range (dtrlim*dt).

    Takes number of initial points to :skip.

    This method is inspired by Trilinos Rythmos::FirstOrderErrorStepControlStrategy.
    """
    if weight is None:
        weight = model.transientvar
    failures = 0
    step = 1
    attempt = 1
    logger = logs.timestepping
    while t < t1:
        # t,x is current solution
        if dt < mindt:
            raise SolverError(
                'transient timestep smaller than minimum timestep %s' % mindt)
        if step >= maxsteps:
            raise SolverError(
                'maximum number of %s timesteps reached' % maxsteps)
        assert dt <= maxdt
        tn = t + dt
        if final_matchstep and tn > t1:
            tn = t1
            dt = t1 - t
        # Calculate predicted solution by forward Euler
        xnp = None
        if skip:
            skip -= 1
        elif xt is not None:
            xnp = x + xt * dt
        # Calculate new solution by backward Euler xt=(xn-x)/dt
        logger.info(
            'Timestep %d (attempt %d) at time %e with dt %e' %
            (step, attempt, t, dt))
        attempt += 1
        try:
            if use_predictor:
                initial_guess = xnp
            else:
                initial_guess = x
            xn = solve(model, initial_guess, params, tconst=-x / dt,
                       tshift=1. / dt, time=tn, solver=solver, **kwargs)
            xtn = (xn - x) / dt
            dt_ = dt
        except SolverError:
            dt *= dtrfailsolve
            logger.info('failure in solve, reducing timestep to %e' % (dt))
            failures += 1
            continue
        if xnp is not None:
            dn = (xn - xnp) / 2
            goalnorm = reltol * np.linalg.norm(xn * weight) + abstol
            dnorm = np.linalg.norm(dn * weight)
            if dnorm > relfail * goalnorm:
                dt *= dtrfaillte
                failures += 1
                logger.info('error too big, reducing timestep to %e' % (dt))
                continue
            e = goalnorm / (dnorm + 1e-100)
            dt = np.clip(dt * np.sqrt(e), min(mindt,
                                              dtrlim[0] * dt), min(dtrlim[1] * dt, maxdt))
        t, x, xt = tn, xn, xtn
        step += 1

        def output():
            d = model.output(t, x, xt, params)
            d['tsadapt.failures'] = failures
            d['tsadapt.dt'] = dt
            d['tsadapt.time'] = t
            return d
        yield (t, dt_, x, xt, output)


def bdf1adapt(*args, **kwargs):
    for t, dt, x, xt, output in bdf1adapt_(*args, **kwargs):
        yield t, x, xt, output

# In[43]:

# <api>


def transientsolve_(model, x0, params, timesteps,
                    solve=solve, mindt=1e-15, force=False, maxsteps=100, solver=None, **kwargs):
    "Return sequence of transient solutions (t,dt,x(t),xt,output) for predefined timesteps"
    x = x0
    logger = logs.timestepping
    assert len(timesteps) - 1 <= maxsteps, 'not enough steps allowed'

    def check_step(step):
        if step > maxsteps:
            raise SolverError(
                'maximum number of %s timesteps reached' % maxsteps)
    step = 0
    for i, dt in enumerate(np.diff(timesteps)):
        check_step(step)
        t0, t1 = timesteps[i:i + 2]
        try:
            logger.info('t=%e dt=%e' % (t0, dt))
            xn = solve(model, x, params, tconst=-x / dt,
                       tshift=1. / dt, time=t1, solver=solver, **kwargs)
            x, xt = xn, (xn - x) / dt
            step += 1
            yield (timesteps[i + 1], dt, x, xt,
                   lambda: model.output(t1, x, xt, params))
        except SolverError:
            if not force:
                raise
            logger.info('timestep failed')
            if 0.5 * (t1 - t0) < mindt:
                raise
            for t, dt, x, xt, output in transientsolve_(
                    model, x, params, [t0, 0.5 * (t0 + t1), t1], maxsteps=np.inf, **kwargs):
                check_step(step)
                step += 1
                yield t, dt, x, xt, output


def transientsolve(*args, **kwargs):
    for t, dt, x, xt, output in transientsolve_(*args, **kwargs):
        yield t, x, xt, output


class ACSolver:
    def __init__(self, model, time, x, xt, params,
                 inparam, outfunc, spsolve=spsolve_scipy, solver=None):
        n = len(x)
        sparams = paramvector([inparam])
        p = sparams.values(params)
        u = np.concatenate([x, xt, p])
        adu = forward.seed_sparse_gradient(u)
        x, xt = adu[:n], adu[n:2 * n]
        p = sparams.asdict(params, adu[2 * n:])
        full_output = {}
        F = model.residuals(
            time,
            x,
            xt,
            p,
            full_output=full_output,
            solver=solver)
        output = outfunc
        J = output(model, time, x, xt, p, full_output)

        def split(a):
            A = a.gradient.tocsr()
            return a.value, (A[:, :n], A[:, n:2 * n], A[:, 2 * n:])
        _, (self.F_x, self.F_xt, F_p) = split(F)
        self.J, (self.J_x, self.J_xt, self.J_p) = split(J)
        self.rhs = -F_p.todense()
        self.spsolve = spsolve
        self.scaling = model.scaling(params)

    def solve(self, omega):
        iw = 1.j * omega
        A = self.F_x + iw * self.F_xt
        X = _matrixsolve(self.spsolve, A, self.rhs, self.scaling)
        J = (self.J_x + 1.j * omega * self.J_xt).dot(X) + self.J_p
        assert J.shape == (1, 1)
        return X, J[0, 0]



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .base import ConservationEquation
from oedes.fvm import FVMPoissonEquation, ElementwiseConvergenceTest
from oedes.ad import dot, sum, getitem
import scipy.constants
from oedes.models import solver
from oedes.utils import WeakList, Funcall

__all__ = ['PoissonEquation', 'Poisson']


class PoissonEquation(ConservationEquation):
    """
    Poisson's equation

    Checkpoint allspecies is run after currents for all species inside the domain are calculated
    Checkpoint alldone is the last before freeing variables
    """

    def __init__(self, *args, **kwargs):
        super(Poisson, self).__init__(*args, **kwargs)
        self.convergenceTest = ElementwiseConvergenceTest(
            atol=1e-15, rtol=1e-7)
        self.defaultBCConvergenceTest = ElementwiseConvergenceTest(
            atol=1e-15, rtol=1e-7)

    def newDiscreteEq(self, builder):
        return builder.newPoissonEquation(
            builder.getMesh(self.mesh), self.name)

    def buildDiscreteEq(self, builder, obj):
        super(Poisson, self).buildDiscreteEq(builder, obj)
        obj.species = WeakList()
        obj.allspecies = Funcall(self.allspecies, obj, depends=[obj.evaluate])
        obj.alldone = Funcall(None, obj, depends=[obj.allspecies])
        obj.ramo_shockley_potentials = dict()
        builder.addEvaluation(obj.allspecies)

    def load(self, ctx, eq):
        super(Poisson, self).load(ctx, eq)
        variables = ctx.varsOf(eq)
        potential = variables['x']
        E = eq.E(potential)
        Et = eq.E(variables['xt'])
        if isinstance(ctx.solver, solver.RamoShockleyCalculation):
            epsilon = scipy.constants.epsilon_0
        else:
            epsilon = scipy.constants.epsilon_0 * \
                ctx.param(eq, 'epsilon_r')
        Ecellv = eq.mesh.cellaveragev(E)
        Ecellm = eq.mesh.magnitudev(Ecellv)
        variables.update(
            Ecellm=Ecellm,
            Ecellv=Ecellv,
            epsilon=epsilon,
            E=E,
            Et=Et,
            potential=potential)

    def evaluate(self, ctx, eq):
        "Part of evaluate dealing with Poisson's equation"
        assert isinstance(eq, FVMPoissonEquation)
        variables = ctx.varsOf(eq)
        total_charge_density = 0
        if not isinstance(ctx.solver, solver.RamoShockleyCalculation):
            for s in eq.species:
                total_charge_density = total_charge_density + \
                    s.ze * ctx.varsOf(s)['c']
        elif ctx.solver.store:
            eq.ramo_shockley_potentials[ctx.solver.boundary_name] = variables['potential']
        epsilon = variables['epsilon']
        faceepsilon = eq.mesh.faceaverage(epsilon)
        potential = variables['potential']
        E = variables['E']
        Et = variables['Et']
        D = eq.displacement(E, faceepsilon)
        Dt = eq.displacement(Et, faceepsilon)
        variables.update(D=D, Dt=Dt)
        ctx.outputCell([eq, 'total_charge_density'],
                       total_charge_density, unit=ctx.units.charge_density)
        ctx.outputCell([eq, 'potential'],
                       potential, unit=ctx.units.potential)
        ctx.outputFace([eq, 'E'], E, unit=ctx.units.electric_field)
        ctx.outputFace([eq, 'Et'], Et)
        ctx.outputFace([eq, 'D'], D)
        ctx.outputFace([eq, 'Dt'], Dt, unit=ctx.units.current_density)
        variables['Jd_boundary'] = dot(eq.mesh.boundary.fluxsum, Dt)
        return self.residuals(ctx, eq, flux=D, source=total_charge_density)

    def allspecies(self, ctx, eq):
        if not ctx.wants_output or ctx.solver.poissonOnly:
            return
        result = dict()
        for name, phi in eq.ramo_shockley_potentials.items():
            w = eq.mesh.facegrad(
                phi) * eq.mesh.faces['dr'] * eq.mesh.faces['surface']
            J = sum(w * ctx.varsOf(eq)['Dt'])
            for s in eq.species:
                svars = ctx.varsOf(s)
                j = ctx.varsOf(s)['j']
                if j is None:
                    continue
                J = J + sum(w * j * s.ze)
            result[name] = J
        ctx.varsOf(eq)['Jramo_shockley'] = result


Poisson = PoissonEquation



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import numpy as np
from .deprecated import BaseModel
from oedes.models.equations import BoltzmannDOS, BandTransport, AdvectionDiffusion, Electroneutrality, Poisson, NontransportedSpecies
from oedes.models.equations.general import species_v_D_charged_from_params
from oedes.models.equations import AppliedVoltage, FermiLevelEqualElectrode
from .source import LangevinRecombination, DirectRecombination, SRH, TrapSource, DirectGeneration
from oedes.fvm import mesh1d


def add_transport(model, mesh, z, name, traps=[],
                  dos_class=BoltzmannDOS, **kwargs):
    """
    Adds transport equation to the model.
    Also adds trap levels and source terms transport<->trap level.
    """

    eq = BandTransport(
        mesh=mesh,
        name=name,
        z=z,
        dos=dos_class(),
        thermal=model.thermal,
        **kwargs)
    eq.bc = [FermiLevelEqualElectrode(boundary)
             for boundary in mesh.boundaries]
    model.species.append(eq)
    basename = name
    for name in traps:
        teq = NontransportedSpecies(mesh, basename + '.' + name, z=z)
        model.species.append(teq)
        model.sources.append(TrapSource(eq, teq))
    return eq


def electronic_device(model, mesh, polarity, ntraps=[], ptraps=[], **kwargs):
    """
    Creates simple electronic device:
    - if Poisson's equation is not present, create default with voltage
    - creates 'hole' if 'p' polarity requested, optionally with corresponding traps
    - creates 'electron' if 'n' polarity requested, optionally with corresponding traps
    - if both polarities are present, adds recombination term
    """
    if model.poisson.mesh is None:
        model.poisson = Poisson(mesh)
        model.poisson.bc = [AppliedVoltage(boundary)
                            for boundary in mesh.boundaries]
    assert polarity in ['n', 'p', 'np', 'pn', ''], "invalid polarity"
    electron, hole = None, None
    if 'n' in polarity:
        electron = add_transport(
            model, mesh, -1, 'electron', traps=ntraps, **kwargs)
    if 'p' in polarity:
        hole = add_transport(model, mesh, +1, 'hole', traps=ptraps, **kwargs)
    if electron is not None and hole is not None:
        r = LangevinRecombination(electron, hole)
        model.sources.append(r)


def _find(model):
    return model.findeq('cation'), model.findeq('anion')


def _initial_salt(model, cinit, nc=1., na=1.):
    "Return initial state of device corresponding to uniformly distributed ions with salt concentration cinit"

    cation, anion = _find(model)
    assert nc > 0 and na > 0 and cation.z > 0 and anion.z < 0
    assert cation.z * nc == -anion.z * na
    x = np.zeros_like(model.X)
    x[cation.idx] = cinit * nc
    x[anion.idx] = cinit * na
    return x


def add_ions(model, mesh, zc=1, za=-1):
    assert zc > 0 and za < 0, "follow standard convention"
    if model.poisson.mesh is None:
        model.poisson = Poisson(mesh)
        model.poisson.bc = [AppliedVoltage(k) for k in mesh.boundaries]
    cation = AdvectionDiffusion(
        mesh=mesh,
        name='cation',
        z=zc,
        thermal=model.thermal,
        v_D=species_v_D_charged_from_params)
    anion = AdvectionDiffusion(
        mesh=mesh,
        name='anion',
        z=za,
        thermal=model.thermal,
        v_D=species_v_D_charged_from_params)
    model.species.extend([cation, anion])
    return cation, anion, lambda *args, **kwargs: _initial_salt(
        model, *args, **kwargs)


electrolyte = add_ions


def holeonly(L, traps=[]):
    model = BaseModel()
    electronic_device(model, mesh1d(L), 'p', ptraps=traps)
    model.setUp()
    return model


def electrononly(L, traps=[]):
    model = BaseModel()
    electronic_device(model, mesh1d(L), 'n', ntraps=traps)
    model.setUp()
    return model


def bulk_heterojunction(model, mesh, ptraps=[], ntraps=[], selective_contacts=False, absorption=lambda x: 1.,
                        langevin_recombination=False, const_recombination=False, srh_recombination=False):
    assert any(
        (langevin_recombination,
         const_recombination,
         srh_recombination))
    if model.poisson.mesh is None:
        model.poisson = Poisson(mesh)
        model.poisson.bc = [AppliedVoltage(k) for k in mesh.boundaries]
    electron = add_transport(model, mesh, -1, 'electron', traps=ntraps)
    hole = add_transport(model, mesh, 1, 'hole', traps=ptraps)
    if selective_contacts:
        electron.bc = [FermiLevelEqualElectrode('electrode1')]
        hole.bc = [FermiLevelEqualElectrode('electrode0')]
    model.sources.append(DirectGeneration([electron, hole], absorption))
    semiconductor = Electroneutrality([electron, hole], name='semiconductor')
    model.other = [semiconductor]
    if langevin_recombination:
        model.sources.append(LangevinRecombination(semiconductor))
    if const_recombination:
        model.sources.append(DirectRecombination(semiconductor))
    if srh_recombination:
        model.sources.append(SRH(semiconductor, name='srh'))
    return model


def bulk_heterojunction_params(
        barrier=0.3, bandgap=1.2, Nc=1e27, Nv=1e27, T=300., epsilon_r=5., mu_e=1e-10, mu_h=1e-10):
    p = {
        'T': T,
        'epsilon_r': epsilon_r,
        'electron.mu': mu_e,
        'hole.mu': mu_h,
        'electron.N0': Nc,
        'hole.N0': Nv,
        'generation.I': 1.,
        'electron.energy': 0.,
        'electrode0.workfunction': bandgap - barrier,
        'hole.energy': -bandgap,
        'electrode1.workfunction': barrier,
        'electrode0.voltage': 0.,
        'electrode1.voltage': 0.
    }
    return p



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import oedes


def test_no_garbage():
    import gc
    gc.collect()
    gc.disable()
    gc.set_debug(gc.DEBUG_SAVEALL)

    def run():
        model, params = oedes.testing.unipolar(100e-9)
        model.setUp()
        c = oedes.context(model)
        c.solve(params)
        c.output()
    run()
    gc.collect()
    assert not gc.garbage
    gc.set_debug(0)
    gc.enable()



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['Funcall', 'FuncallGraph']

from collections import defaultdict
import weakref


class Funcall(object):
    def __init__(self, func, obj, depends=()):
        self.func = func
        self._obj = weakref.ref(obj)
        self._depends = list(depends)

    def depends(self, f):
        assert isinstance(f, Funcall)
        #assert f not in self._depends
        self._depends.append(f)

    def __hash__(self):
        return hash(id(self))

    def __eq__(self, other):
        return self is other

    def __call__(self, ctx):
        return self.func(ctx, self.obj())

    def obj(self):
        obj = self._obj()
        assert obj is not None
        return obj


class FuncallGraph(object):
    def __init__(self):
        self.evaluations = []

    def add(self, func):
        self.evaluations.append(func)

    def plan(self, kill_func=None):
        evaluations = list(self.evaluations)
        rdeps = defaultdict(list)
        queue = list()

        # Build reverse dependence graph `rdeps`, and set of all calculations
        # `known`
        known = set(self.evaluations)
        while evaluations:
            e = evaluations.pop()
            assert isinstance(e, Funcall)
            for d in e._depends:
                assert isinstance(d, Funcall)
                rdeps[d].append(e)
                # if dependence was not added explicitely by calling `add`
                # method, add it here
                if d not in known:
                    known.add(d)
                    evaluations.append(d)
            if not e._depends:
                queue.append(e)

        waitcount = dict((f, len(f._depends)) for f in known)
        usecount_byobj = defaultdict(int)
        for f in known:
            for d in f._depends:
                usecount_byobj[id(d.obj())] += 1
        while queue:
            t = queue.pop()
            if t.func is not None:  # do not yield checkpoints where fun is None
                yield t
            if not usecount_byobj[id(t.obj())]:
                yield kill_func(t.obj())
            for d in t._depends:
                obj = d._obj()
                assert usecount_byobj[id(obj)] > 0
                usecount_byobj[id(obj)] -= 1
                if not usecount_byobj[id(obj)]:
                    if kill_func is not None:
                        yield kill_func(obj)
            for r in rdeps[t]:
                assert waitcount[r] >= 1
                waitcount[r] -= 1
                if waitcount[r] == 0:
                    queue.append(r)
        assert not any(waitcount.values(
        )), 'unfinished tasks left: not added but registered as dependencies or dependence cycles'



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .source import _Recombination
from oedes.ad import exp

__all__ = ['SRH']


class SRH(_Recombination):
    "SRH (steady-state) term"

    def __init__(self, *args, **kwargs):
        name = kwargs.pop('name', 'srh')
        super(SRH, self).__init__(*args, name=name, **kwargs)

    def evaluate(self, ctx, eq):
        assert eq.hole_eq.thermal is eq.electron_eq.thermal
        if ctx.solver.poissonOnly:
            return
        Vt = ctx.varsOf(eq.electron_eq.thermal)['Vt']
        Et = ctx.param(eq, 'energy')
        ni = ctx.param(eq.electron_eq, 'N0') * \
            exp((Et - ctx.param(eq.electron_eq, 'energy')) / Vt)
        pi = ctx.param(eq.hole_eq, 'N0') * \
            exp((ctx.param(eq.hole_eq, 'energy') - Et) / Vt)
        Cn = ctx.param(eq.electron_eq, self.name, 'trate')
        Cp = ctx.param(eq.hole_eq, self.name, 'trate')
        n = ctx.varsOf(eq.electron_eq)['c']
        p = ctx.varsOf(eq.hole_eq)['c']
        g = Cn * Cp * ctx.param(eq, 'N0') * \
            (n * p - ni * pi) / (Cn * (n + ni) + Cp * (p + pi))
        ctx.outputCell([eq, 'G'],  g, mesh=eq.electron_eq.mesh,
                       unit=ctx.units.dconcentration_dt)
        self.add(ctx, g, minus=[eq.electron_eq, eq.hole_eq])



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .band import MobilityModel
from oedes.ad import where, exp, sqrt, log

__all__ = ['FrenkelPooleMobility']


class FrenkelPooleMobility(MobilityModel):
    def mobility(self, parent, ctx, eq):
        mu0 = ctx.param(eq, 'mu0')
        gamma = ctx.param(eq, 'gamma')
        v = gamma * sqrt(ctx.varsOf(eq.poisson)['Ecellm'] + 1e-10)
        v_max = log(1e10)
        v_min = -v_max
        v = where(v < v_max, v, v_max)
        v = where(v > v_min, v, v_min)
        mu = mu0 * exp(v)
        mu_face = eq.mesh.faceaverage(mu)
        ctx.varsOf(eq)['mu_face'] = mu_face
        ctx.varsOf(eq)['mu_cell'] = mu



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.models.equations import DOS, MobilityModel
from oedes.utils import SubEquation
from oedes.ad import where, branch, getitem, nvalue, log
import oedes.functions.egdm as _egdm
import oedes.functions.gdos as _gdos
from oedes.functions.physics import ThermalVoltage
import numpy as np
from oedes import logs


def egdm_params_simple(params, prefix, nsigma, a, mu0t):
    params[prefix + '.N0'] = _egdm.atoN0(a)
    params[
        prefix + '.sigma'] = _egdm.nsigma_inverse(nsigma, ThermalVoltage(params['T']))
    params[prefix + '.mu0'] = _egdm.mu0t_inverse(nsigma, mu0t)


class DisorderFromParams(SubEquation):
    def disorder_parameters(self, parent, ctx, eq):
        info = dict(sigma=ctx.param(eq, 'sigma'),
                    mu0=ctx.param(eq, 'mu0'),
                    Vt=ctx.varsOf(eq.thermal)['Vt'],
                    N0=ctx.param(eq, 'N0'))
        info['nsigma'] = _egdm.nsigma(sigma=info['sigma'], Vt=info['Vt'])
        return info


class GaussianDOS(DOS):
    # it is allowed to turn off any enhancement factor
    use_g3 = True
    # to avoid numeric problems, c is clipped to more than c_eps
    c_eps = 1e-30
    c_max = 0.9999
    # custom limits are allowed
    g3_max_c = _egdm.g3_max_c
    # custom gdos impl is allowed, but must be compatible with
    # limits set here
    gdosImpl = _gdos.defaultImpl

    # To avoid numeric problems, nsigma is clipped to these values.
    # This can be useful in initial Newton iterations with
    # thermal coupling when temperature can be weird, and so nsigma.
    # However, final solution should not be affected by that.
    # Keep in mind that the model is for nsigma=3,4,5,6. nsigma<=1
    # will cause NaN in g1, and nsigma>>6 is problematic for Gauss-
    # Fermi integral.
    nsigma_min = 2.
    nsigma_max = 7.

    logger = logs.models.getChild('GaussianDOS')

    def __init__(self, impl=None, disorder=None):
        super(GaussianDOS, self).__init__()
        if impl is None:
            impl = _gdos.defaultImpl
        self.impl = impl
        if disorder is None:
            disorder = DisorderFromParams()
        self.disorder = disorder

    def _load(self, ctx, eq):
        variables = ctx.varsOf(eq)
        if 'gdos_data' in variables:
            return variables['gdos_data']
        info = self.disorder.disorder_parameters(self, ctx, eq)
        nsigma = info['nsigma']
        nsigma = where(nsigma > self.nsigma_min, nsigma, self.nsigma_min)
        nsigma = where(nsigma < self.nsigma_max, nsigma, self.nsigma_max)
        if np.any(nsigma != info['nsigma']):
            self.logger.info(
                'nsigma clipped to ensure numerical stability: probably not calculating what you want. This is OK in initial Newton iterations, but not in the final one')
        N0 = info['N0']
        c = ctx.varsOf(eq)['c'] / N0
        assert self.c_eps >= self.impl.I_min
        assert self.c_max <= self.impl.I_max
        c = where(c > self.c_eps, c, self.c_eps)
        c = where(c < self.c_max, c, self.c_max)
        b = self.impl.b(np.sqrt(2.) * nsigma, c)
        info.update(b=b, c=c, nsigma=nsigma)
        variables['gdos_data'] = info
        return info

    def Ef(self, ctx, eq):
        info = self._load(ctx, eq)
        return -info['b'] * info['Vt']

    def Ef_correction(self, ctx, eq):
        info = self._load(ctx, eq)
        return (-info['b'] - log(info['c'])) * info['Vt']

    def c(self, ctx, eq,  Ef, numeric_parameters=False):
        info = self._load(ctx, eq)
        N0 = info['N0']
        nsigma = info['nsigma']
        Vt = info['Vt']
        if numeric_parameters:
            N0 = nvalue(N0)
            nsigma = nvalue(nsigma)
            Vt = nvalue(Vt)
        return N0 * self.impl.I(np.sqrt(2.) * nsigma, b=-Ef / info['Vt'])

    def D_mu(self, parent, ctx, eq):
        info = self._load(ctx, eq)
        assert self.g3_max_c <= self.gdosImpl.I_max
        assert self.c_eps >= self.gdosImpl.I_min
        assert self.nsigma_max * np.sqrt(2.) <= self.gdosImpl.a_max
        c = info['c']
        nsigma = info['nsigma']
        inlimit_g3 = c < self.g3_max_c

        def branch_if_inlimit(ix):
            return getitem(info['b'], ix)

        def branch_cut_limit(ix):
            return self.impl.b(
                np.sqrt(2.) * getitem(nsigma, ix), self.g3_max_c)
        b_g3 = branch(inlimit_g3, branch_if_inlimit, branch_cut_limit)
        c_g3 = where(inlimit_g3, c, self.g3_max_c)
        g3 = _egdm.g3(nsigma, c_g3, impl=self.impl, b=b_g3)
        return eq.mesh.faceaverage(g3*info['Vt'])


class EGDMMobility(MobilityModel):
    g1_max_c = _egdm.g1_max_c
    g2_max_En = _egdm.g2_max_En

    def __init__(self, use_g1=True, use_g2=True):
        super(EGDMMobility, self).__init__()
        self.use_g1 = use_g1
        self.use_g2 = use_g2

    def _load(self, ctx, eq):
        return ctx.varsOf(eq)['gdos_data']

    def mobility(self, parent, ctx, eq):
        info = self._load(ctx, eq)
        a = _egdm.N0toa(info['N0'])
        nsigma = info['nsigma']
        sigma = info['sigma']
        mu_cell = _egdm.mu0t(nsigma, info['mu0'])
        c = info['c']
        if self.use_g1:
            mu_cell = mu_cell * \
                _egdm.g1(nsigma, where(c < self.g1_max_c, c, self.g1_max_c))
        mu_face = eq.mesh.faceaverage(mu_cell)
        E = ctx.varsOf(eq.poisson)['E']
        if self.use_g2:
            En = (E * a) / sigma
            En2 = En**2
            En2_max = self.g2_max_En**2
            mu_face = mu_face * \
                _egdm.g2_En2(nsigma, where(En2 < En2_max, En2, En2_max))
        ctx.varsOf(eq).update(
            mu_face=mu_face,
            mu_cell=eq.mesh.cellaveragev(mu_face))



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# Basic equations
#
# Using vertex centered finite volume method to discretize the equations.
# $\frac{du}{dt} = S + \nabla \cdot F $
# by discretized applying Gauss theorem
# $\int_V \frac{du}{dt}dV = \int_V S dV + \int_\Omega F d\Omega$
# where $V$ denotes cell volume, and $\Omega$ denotes cell boundary.
#
# In discretized geometry
# $V \frac{dc_i}{dt} = V s_i + \sum_k A_{ik} f_{k}$
# This is conveniently expressed by Equation.conservation
# $\frac{d\mathbf{c}}{dt} - \mathbf{s} - V^-1 \mathbf{A} \mathbf{f}=F_{internal}$
# where $V^-1$ is diagonal matrix of cell volumes. A is called fluxsum in the code.
#
# At boundary, values calculated by above equation are replaced by
# $b(V^-1 \mathbf{A} \mathbf{f},...)=F_{boundary}$
#
# Final residuals are given as $F_{internal}$ for internal points, and
# $F_{boundary}$ for boundary cells.

from oedes import ad
import numpy as np
from .discrete import *


def _boundaryIdx(eq, boundary):
    return eq.mesh.boundary.idx[boundary['bidx']]


def _boundaryDof(eq, boundary):
    return eq.idx[_boundaryIdx(eq, boundary)]


class FVMBoundaryEquation(DiscreteEquation):
    """
    Cell-center finite volume boundary equation

    Members:
    ---------------------------
    owner_eq : FVMConservationEquation
        Equation this boundary condition applies to
    other_eq_lookup : callable(FVMBuilder) -> FVMConservationEquation
        For internal interfaces: find equation on the other side of interface
    name : str
        Name of this interface
    other_eq : FVMConservationEquation
        For internal interface, after setUp : equation on the other side of the interface
    boundary :
        result of mesh.getBoundary or mesh.getSharedBoundary
    other_boundary :
        For internal interfaces: result of mesh.getSharedBoundary, applies to the other side of the interface
    conservation_to_dof : integer array
        For internal interfaces: destination cell for conservation
    """

    def __init__(self, owner_eq, name, other_eq_lookup=None):
        super(FVMBoundaryEquation, self).__init__()
        self.owner_eq = owner_eq
        self.other_eq_lookup = other_eq_lookup
        self.name = name

    def _getDof(self):
        return _boundaryDof(self.owner_eq, self.boundary)

    def _getIdx(self):
        return _boundaryIdx(self.owner_eq, self.boundary)

    def _getOtherIdx(self):
        return _boundaryIdx(self.other_eq, self.other_boundary)

    def init(self, builder):
        if self.other_eq_lookup is None:
            self.other_eq = None
            self.boundary = self.owner_eq.mesh.getBoundary(self.name)
            self.other_boundary = None
            self.conservation_to_dof = _boundaryDof(
                self.owner_eq, self.boundary)
        else:
            self.other_eq = self.other_eq_lookup(builder)
            self.boundary, self.other_boundary = self.owner_eq.mesh.getSharedBoundary(
                self.other_eq.mesh, self.name)
            self.conservation_to_dof = _boundaryDof(
                self.other_eq, self.other_boundary)
        assert self.owner_eq.bc_dof_is_free[self.boundary['bidx']].all(
        ), 'not overriding BCs'
        self.owner_eq.bc_dof_is_free[self.boundary['bidx']] = False

    def ndof(self):
        return 0

    def testConverged(self, F, x, dx, report):
        return self.convergenceTest.testBoundary(
            self.owner_eq, self, F, x, dx, report)


BoundaryEquation = FVMBoundaryEquation


class FVMConservationEquation(DiscreteEquation):
    """
    Basic equation defining one degree of freedom per cell, solved using vertex centered finite volume

    Properties
    ----------

    transientvar : float
        one if pseudotransient techniques should be applied to this equation, zero if equation is time-independent
    idx : array of int
        indices of degrees of freedom in solution vector corresponding to this equation
    convergenceTest : ConvergenceTest
        convergence criteria for this equation
    prefix : str
        prefix of this equation used for parameters and output
    boundary_labels : integer array
        For internal interfaces: labels of merged interfacial cells
    """

    def __init__(self, mesh, name):
        """Abstract block of equation, which introduces a field into equation system on :mesh:
        Keeps track of:
        indices of belonging to the field :idx:, x[idx[i]] is field in cell i
        :name:, which currently is the same as prefix
        :boundary: cells
        """
        super(FVMConservationEquation, self).__init__()
        self.mesh = mesh
        self.name = name
        self.bc = []

        self.bc_dof_is_free = np.ones_like(mesh.boundary.idx, dtype=np.bool)
        self.boundary_labels = None

    def residuals(self, part, facefluxes, celltransient=0., cellsource=0.):
        """
        Calculate FVM residuals

        Parameters:
        -----------
        facefluxes: vector or None
            fluxes per unit of surface or None
        celltransient: float or vector
            transient term per unit of volume
        cellsource: float or vector
            source term per unit of volume


        Returns: tuple (residuals,FdS)
        --------
        residuals:
            residuals for each cell, per unit of volume
        FdS:
            integral FdS per cell
        """

        assert part is self.mesh.internal or part is self.mesh.boundary
        idx = part.idx
        if facefluxes is None:
            FdS = 0.
        else:
            FdS = ad.dot(part.fluxsum, facefluxes)
        if not ad.isscalar(celltransient):
            celltransient = celltransient[idx]
        if not ad.isscalar(cellsource):
            cellsource = cellsource[idx]
        return self.idx[idx], -FdS / \
            part.cells['volume'] + celltransient - cellsource

    def ndof(self):
        return self.mesh.ncells

    def identity(self, x):
        return self.idx, x - ad.nvalue(x)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['BandTransport', 'MobilityModel', 'MobilityFromParams']

from .boltzmann import BoltzmannDOS
from .transport import TransportedSpecies
from .dos import WithDOS
from oedes.utils import SubEquation
from oedes.ad import where, exp, sqrt, log


class MobilityModel(SubEquation):
    def mobility(self, parent, ctx, parent_eq):
        "Sets mu_face, mu_cell variables. Normally mu_face is used for transport, mu_cell for some other calculations (ie. recombination)"
        raise NotImplementedError()


class MobilityFromParams(MobilityModel):
    def mobility(self, parent, ctx, eq):
        mu = ctx.param(eq, 'mu')
        ctx.varsOf(eq).update(mu_face=mu, mu_cell=mu)


class FrenkelPooleMobility(MobilityModel):
    def mobility(self, parent, ctx, eq):
        mu0 = ctx.param(eq, 'mu0')
        gamma = ctx.param(eq, 'gamma')
        v = gamma * sqrt(ctx.varsOf(eq.poisson)['Ecellm'] + 1e-10)
        v_max = log(1e10)
        v_min = -v_max
        v = where(v < v_max, v, v_max)
        v = where(v > v_min, v, v_min)
        mu = mu0 * exp(v)
        mu_face = eq.mesh.faceaverage(mu)
        ctx.varsOf(eq)['mu_face'] = mu_face
        ctx.varsOf(eq)['mu_cell'] = mu


class BandTransport(WithDOS, TransportedSpecies):
    alternative = False

    def __init__(self, mesh=None, name=None, z=1, poisson=None,
                 thermal=None, dos=None, mobility_model=None):
        assert z in [1, -1]
        super(
            BandTransport,
            self).__init__(
            mesh=mesh,
            name=name,
            z=z,
            poisson=poisson,
            thermal=thermal)
        if dos is None:
            dos = BoltzmannDOS()
        if mobility_model is None:
            mobility_model = MobilityFromParams()
        self.dos = dos
        self.mobility_model = mobility_model

    def evaluate_mobility(self, ctx, eq):
        self.mobility_model.mobility(self, ctx, eq)

    def initDiscreteEq(self, builder, obj):
        super(BandTransport, self).initDiscreteEq(builder, obj)
        self.mobility_model.subInit(self, builder, obj)

    def evaluate_fluxes(self, ctx, eq):
        mu = ctx.varsOf(eq)['mu_face']
        if self.alternative:
            Ef_correction = self.dos.Ef_correction(ctx, eq)
            v = eq.v(
                mu,
                ctx.varsOf(
                    eq.poisson)['E'] -
                eq.mesh.facegrad(Ef_correction) *
                eq.z)
            D = eq.D(mu, ctx.varsOf(eq.thermal)['Vt'])
        else:
            D_mu = self.dos.D_mu(self, ctx, eq)
            v = eq.v(mu, ctx.varsOf(eq.poisson)['E'])
            D = eq.D(mu, D_mu)
        return self.calculate_fluxes(ctx, eq, v, D)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from .base import BoundaryEquation
from .poisson import PoissonEquation
from oedes.utils import Funcall
from oedes import functions
from oedes.ad import where, getitem, sum
from oedes.models import solver

__all__ = [
    'DirichletBC',
    'DirichletBCWithFunction',
    'DirichletFromParams',
    'Zero',
    'DirichletFromParams',
    'Internal',
    'Equal',
    'FermiLevelEqual',
    'FermiLevelEqualElectrode',
    'AppliedVoltage',
    'LocalThermalEquilibrium',
    'LocalThermalEquilibriumContact']


class DirichletBC(BoundaryEquation):
    def evaluate(self, ctx, eq):
        if ctx.solver.poissonOnly and not isinstance(
                self.owner_eq, PoissonEquation):
            return
        yield eq._getDof(), ctx.varsOf(eq.owner_eq)['x'][eq._getIdx()] - self.value(ctx, eq)

    def value(self, ctx, eq):
        raise NotImplementedError()


class DirichletBCWithFunction(DirichletBC):
    def __init__(self, name, function):
        super(DirichletBCWithFunction, self).__init__(name)
        self.function = function

    def initDiscreteEq(self, builder, obj):
        super(DirichletBCWithFunction, self).initDiscreteEq(builder, obj)
        self.function.subInit(self, builder, obj)

    def value(self, ctx, eq):
        return self.function.value(self, ctx, eq)


class AppliedVoltage(DirichletBC):
    def __init__(self, *args, **kwargs):
        calculate_current = kwargs.pop('calculate_current', False)
        super(AppliedVoltage, self).__init__(*args, **kwargs)
        self.calculate_current = calculate_current

    def buildDiscreteEq(self, builder, obj):
        super(AppliedVoltage, self).buildDiscreteEq(builder, obj)

        # register current calculation self.allspecies, to be evaluated between
        # allspecies and alldone checkpoints of owner equation
        obj.allspecies = Funcall(self.allspecies, obj)
        obj.allspecies.depends(obj.owner_eq.allspecies)
        builder.addEvaluation(obj.allspecies)
        obj.owner_eq.alldone.depends(obj.allspecies)
        obj.rs_terminal_name = self.name

    def value(self, ctx, eq):
        if isinstance(ctx.solver, solver.RamoShockleyCalculation):
            return where(self.name == ctx.solver.boundary_name, 1, 0)
        return self.potential(ctx, eq)

    def potential(self, ctx, eq):
        return self.voltage(ctx, eq) - ctx.param(eq, 'workfunction')

    def voltage(self, ctx, eq):
        return ctx.param(eq, 'voltage')

    def allspecies(self, ctx, eq):
        if not ctx.wants_output or ctx.solver.poissonOnly:
            return
        if not self.calculate_current:
            return
        bidx = eq.boundary['bidx']
        J = sum(ctx.varsOf(eq.owner_eq)['Jd_boundary'][bidx])
        for s in eq.owner_eq.species:
            J = J + sum(ctx.varsOf(s)['total_boundary_FdS'][bidx]*s.ze)
        ctx.output([eq, 'Jboundary'], J)


class Zero(DirichletBC):
    def value(self, ctx, eq):
        return 0


class DirichletFromParams(DirichletBC):
    def value(self, ctx, eq):
        return ctx.param(eq.owner_eq, self.name)


class Internal(DirichletBC):
    def __init__(self, other_eq, name):
        super(Internal, self).__init__(name)
        self.name = name
        self.eq_from = other_eq

    def newDiscreteEq(self, builder):
        return builder.newDirichletBC(builder.get(self.owner_eq), self.name,
                                      other_eq_lookup=lambda builder: builder.get(self.eq_from))

    def initDiscreteEq(self, builder, obj):
        super(Internal, self).initDiscreteEq(builder, obj)
        obj.evaluate.depends(builder.get(self.eq_from).load)

        # make sure that this BC will be evaluated before variables of eq_from
        # are freed
        builder.get(self.eq_from).alldone.depends(obj.evaluate)


class Equal(Internal):
    def value(self, ctx, eq):
        return ctx.varsOf(eq.other_eq)['x'][eq._getOtherIdx()]


class FermiLevelEqual(Internal):
    def value(self, ctx, eq):
        if ctx.solver.poissonOnly:
            assert not isinstance(eq, PoissonEquation)
            return
        Effrom = ctx.varsOf(eq.other_eq)['Ef'][eq._getOtherIdx()]
        return self.owner_eq.dos.concentration(
            ctx, eq.owner_eq, eq._getIdx(), Effrom)


class FermiLevelEqualElectrode(DirichletBC):
    F_eps = 1e-10

    def __init__(self, name, image_force=False, **kwargs):
        super(FermiLevelEqualElectrode, self).__init__(name, **kwargs)
        self.image_force = image_force

    def image_correction(self, ctx, eq, ixto):
        if not self.image_force:
            return 0.
        n = eq.mesh.boundaries[self.name]['normal']
        F = eq.z * eq.mesh.dotv(ctx.varsOf(eq.poisson)['Ecellv'][ixto], n)
        F = where(F > self.F_eps, F, self.F_eps)
        return -eq.z * \
            functions.EmtageODwyerBarrierLowering(
                F, getitem(ctx.varsOf(eq.poisson)['epsilon'], ixto))

    def value(self, ctx, eq):
        dos = self.owner_eq.dos
        ixto = eq._getIdx()
        Ef_electrode = -ctx.param(eq, 'voltage') + \
            self.image_correction(ctx, eq.owner_eq, ixto)
        c_electrode = dos.concentration(ctx, eq.owner_eq, ixto, Ef_electrode)
        return c_electrode


LocalThermalEquilibrium = FermiLevelEqual
LocalThermalEquilibriumContact = FermiLevelEqualElectrode



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['WeakList']

import weakref


def _unwrap(item):
    assert isinstance(item, weakref.ReferenceType)
    return item()


def _wrap(obj):
    assert not isinstance(obj, weakref.ReferenceType)
    return weakref.ref(obj)


class WeakList(list):
    """List of weak references emulating normal list"""

    def __init__(self, items=None):
        if items is None:
            items = []
        super(WeakList, self).__init__(map(_wrap, items))

    def __contains__(self, obj):
        return super(WeakList, self).__contains__(self, _wrap(obj))

    def __getitem__(self, index):
        result = super(WeakList, self).__getitem__(index)
        if isinstance(result, list):
            return list(map(_unwrap, result))
        return _unwrap(result)

    def __setitem__(self, index, item):
        if isinstance(index, slice):
            super(WeakList, self).__setitem__(index, map(_wrap, item))
        else:
            return _wrap(item)

    def __iter__(self):
        itr = super(WeakList, self).__iter__()
        return iter(_unwrap(x) for x in itr)

    def __reversed__(self):
        result = type(self)(self)
        result.reverse()
        return result

    def __iadd__(self, other):
        self.extend(WeakList(other))
        return self

    def __eq__(self, other):
        return list(self) == other

    def __neq__(self, other):
        return list(self) != other

    def __lt__(self, other):
        return list(self) < other

    def __le__(self, other):
        return list(self) <= other

    def __gt__(self, other):
        return list(self) > other

    def __ge__(self, other):
        return list(self) >= other

    def append(self, item):
        super(WeakList, self).append(_wrap(item))

    def remove(self, item):
        super(WeakList, self).remove(_wrap(item))

    def index(self, item):
        return super(WeakList, self).index(_wrap(item))

    def count(self, item):
        return super(WeakList, self).count(_wrap(item))

    def pop(self, index=-1):
        return _unwrap(super(WeakList, self).pop(index))

    def insert(self, index, item):
        return super(WeakList, self).insert(index, _wrap(item))

    def extend(self, items):
        return super(WeakList, self).extend(map(_wrap, items))

    def sort(self, *args, **kwargs):
        result = list(map(_unwrap, self)).sort(*args, **kwargs)
        self.clear()
        self.extend(result)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.utils import Calculation
import scipy.constants
from oedes.ad import sum
from oedes.models import solver
from collections import defaultdict
from oedes.utils import ArgStack
from oedes.models.solver import RamoShockleyCalculation
from oedes import solve
import numpy as np

__all__ = ['ConstTemperature', 'RamoShockleyCurrentCalculation']


class ConstTemperature(Calculation):
    def load(self, ctx, eq):
        if isinstance(ctx.solver, solver.RamoShockleyCalculation):
            return
        T = ctx.param(eq, 'T')
        Vt = T * scipy.constants.Boltzmann / scipy.constants.elementary_charge
        ctx.newVars(eq).update(dict(T=T, Vt=Vt))


class RamoShockleyCurrentCalculation(Calculation):
    def __init__(self, poisson_eqs, name=None):
        super(RamoShockleyCurrentCalculation, self).__init__(name)
        self.poisson_eqs = poisson_eqs

    def initDiscreteEq(self, builder, obj):
        super(
            RamoShockleyCurrentCalculation,
            self).initDiscreteEq(
            builder,
            obj)
        obj.poisson_eqs = tuple(map(builder.get, self.poisson_eqs))
        for p in obj.poisson_eqs:
            obj.evaluate.depends(p.allspecies)
        builder.registerFinalizer(self.generateTestfunctions)

    def evaluate(self, ctx, eq):
        if not ctx.wants_output:
            return
        if ctx.solver.poissonOnly:
            return
        d = defaultdict(int)
        default_electrodes = set()
        for p in eq.poisson_eqs:
            Jdict = ctx.varsOf(p)['Jramo_shockley']
            for k in Jdict.keys():
                d[k] = d[k] + Jdict[k]
            if p.mesh.default_electrode is not None:
                default_electrodes.add(p.mesh.default_electrode)
        for k, v in d.items():
            ctx.output([eq, k, 'J'], v)
        if len(eq.poisson_eqs) == 1:
            if default_electrodes and any(k in d for k in default_electrodes):
                Jdefault = 0
                for k in default_electrodes:
                    if k in d:
                        Jdefault = Jdefault + d[k]
                ctx.output([eq, 'J'], Jdefault)

    @classmethod
    def generateTestfunctions(
            cls, builder, model, dtype=np.double, solve_kwargs=None):
        if solve_kwargs is None:
            solve_kwargs = dict()
        rs_boundaries = set()
        for eq in builder.equations:
            if hasattr(eq, 'rs_terminal_name'):
                rs_boundaries.add(eq.rs_terminal_name)
        params = dict()
        for b in rs_boundaries:
            s = RamoShockleyCalculation(b, store=False)
            X = np.asarray(model.X, dtype=dtype)
            x = solve(model, X, params, niter=1, solver=s, **solve_kwargs)
            s = RamoShockleyCalculation(b, store=True)
            model.output(
                0,
                x,
                np.zeros_like(x),
                params,
                solver=s,
                **solve_kwargs)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .cell import FVMConservationEquation


class FVMPoissonEquation(FVMConservationEquation):
    "Poisson equation"

    def __init__(self, mesh, name='poisson'):
        super(FVMPoissonEquation, self).__init__(mesh, name=name)
        self.transientvar = 0.

    def E(self, phi):
        "Return magnitude of electric field at faces"
        return -self.mesh.facegrad(phi)

    def displacement(self, E, epsilon):
        return E * epsilon

    def scaling(self, xscaling, fscaling):
        lunit = 1e-9
        fscaling[self.idx] = lunit

    faceflux = displacement



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.utils import FuncallGraph

__all__ = ['BuilderData']


class BuilderData(object):
    """
    Data used during construction of discrete model

    The discretization process proceeds as follows:

    1. For each calculation, discrete object is build. At this stage,
    calculations add added to graph using `addEvaluation`, and discrete
    objects add using `add`. Initialization that must be done on
    complete system is registered by `addInitializer`.

    2. After all objects are built, initializers registered using `addInitializer`
    are called. Also, `init` methods in discrete objects are called (`initEquations` functions)

    3. Evaluator is constructed using attributes.

    4. After evaluator is constructed, final initializers registered using `registerFinalizer`
    are called with (builder, model, *args)

    Attributes
    ----------
    equations : list
        All discrete objects
    fungraph : FuncallGraph
        Evaluation graph
    """

    def __init__(self):
        self.equations = []
        self.key_to_eq = dict()
        self.init_list = []
        self.finalizers = set()
        self.fungraph = FuncallGraph()

    def getMesh(self, key):
        "Return concrete mesh object for mesh handle"
        raise NotImplementedError()

    def get(self, owner):
        "Return discrete object for calculation object"
        return self.key_to_eq[id(owner)]

    def addInitializer(self, func, *args):
        """
        Add initializer

        func(*args) will be called after discrete objects are constructed for all calculations
        """
        self.init_list.append((func, args))

    def registerFinalizer(self, func, *args):
        """
        Register final initializer

        func(evaluator, *args) will be called after evaluator is constructed
        """
        self.finalizers.add((func, args))

    def addEvaluation(self, func):
        "Add function evaluation to discrete model"
        self.fungraph.add(func)

    def add(self, owner, eq):
        """
        Add discrete object `eq` corresponding to `owner`

        `eq` can is retreived by calling ``self.get(owner)``. Only one discrete equation can
        exist for `owner`
        """
        assert id(owner) not in self.key_to_eq
        self.equations.append(eq)
        self.key_to_eq[id(owner)] = eq

    def initEquations(self):
        "Run late initialization"
        for func, args in self.init_list:
            func(self, *args)
        for eq in self.equations:
            eq.init(self)

    def finalize(self, evaluator):
        for func, args in self.finalizers:
            func(self, evaluator, *args)



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__all__ = ['BoltzmannDOS']

from oedes.ad import where, nvalue
import logging
import numpy as np
from oedes import logs
from .dos import DOS


class BoltzmannDOS(DOS):
    c_eps = 1e-30
    c_limit = True
    Ef_max = None
    logger = logs.models.getChild('BoltzmannDOS')

    def N0(self, ctx, eq):
        return ctx.param(eq, 'N0')

    def Ef(self, ctx, eq):
        c = ctx.varsOf(eq)['c']
        c_raw = c
        c = where(c > self.c_eps, c, self.c_eps)  # avoid NaN
        if self.logger.isEnabledFor(logging.INFO):
            if np.any(nvalue(c_raw) != nvalue(c)):
                self.logger.info(
                    'Ef(%r): clipping c<%r, min(c)=%r' %
                    (eq.prefix, self.c_eps, np.amin(
                        nvalue(c_raw))))
        N0 = self.N0(ctx, eq)
        return ctx.varsOf(eq.thermal)['Vt'] * np.log(c / N0)

    def Ef_correction(self, ctx, eq):
        return 0.

    def c(self, ctx, eq, Ef, numeric_parameters=False):
        Ef_raw = Ef
        if self.Ef_max is not None:
            Ef = where(Ef < self.Ef_max, Ef, self.Ef_max)
            if self.logger.isEnabledFor(logging.INFO):
                if np.any(nvalue(Ef) != nvalue(Ef_raw)):
                    self.logger.info(
                        'c(%r): clipping Ef>%r, max(Ef)=%r' %
                        (eq.prefix, self.Ef_max, np.amax(
                            nvalue(Ef_raw))))
        N0 = self.N0(ctx, eq)
        if numeric_parameters:
            N0 = nvalue(N0)
        Vt = ctx.varsOf(eq.thermal)['Vt']
        if numeric_parameters:
            Vt = nvalue(Vt)
        v = Ef / Vt
        if self.c_limit:
            v_raw = v
            v = where(v <= 0., v, 0.)
            if self.logger.isEnabledFor(logging.INFO):
                if np.any(nvalue(v_raw) != nvalue(v)):
                    self.logger.info(
                        'c(%r): clipping Ef/kT>0, max(Ef/kT)=%r' %
                        (eq.prefix, np.amax(
                            nvalue(v_raw))))
        c = np.exp(v) * N0
        return c

    def D_mu(self, parent_eq, ctx, eq):
        return ctx.varsOf(eq.thermal)['Vt']



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from oedes.utils import Funcall
from .conserved import ConservedSpecies

__all__ = ['TransportedSpecies']


class TransportedSpecies(ConservedSpecies):
    def evaluate_fluxes(self, ctx, eq):
        v_D = ctx.varsOf(eq)['v_D']
        if v_D is None:
            return None
        return self.calculate_fluxes(ctx, eq, *v_D)

    def buildDiscreteEq(self, builder, obj):
        super(TransportedSpecies, self).buildDiscreteEq(builder, obj)
        obj.mobility = Funcall(self.mobility, obj, depends=[obj.load])
        obj.evaluate.depends(obj.mobility)
        builder.addEvaluation(obj.mobility)

    def mobility(self, ctx, eq):
        if ctx.solver.poissonOnly:
            return
        self.evaluate_mobility(ctx, eq)

    def evaluate_mobility(self, ctx, eq):
        raise NotImplementedError()



# -*- coding: utf-8; -*-
#
# oedes - organic electronic device simulator
# Copyright (C) 2017-2018 Marek Zdzislaw Szymanski (marek@marekszymanski.com)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License, version 3,
# as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from .base import ConservationEquation
from .charged import ChargedSpecies
from oedes.fvm import ElementwiseConvergenceTest

__all__ = ['ConservedSpecies', 'NontransportedSpecies']


class ConservedSpecies(ChargedSpecies, ConservationEquation):
    """
    Species with charge, and satisfying conservation equation

    Outputs:
    --------
    j, jdrift, jdiff : face variable
        total flux density, drift part of flux density, diffusion part of flux density  [ 1/(m^2 s)]
    J, Jdrift, Jdiff : face variable
        total current density, drift part of current density, diffusion part of current density  [ 1/(m^2 s)]
    """

    def __init__(self, *args, **kwargs):
        super(ConservedSpecies, self).__init__(*args, **kwargs)
        self.convergenceTest = ElementwiseConvergenceTest(
            atol=5e7, rtol=1e-7, eps=1e-10)
        self.defaultBCConvergenceTest = ElementwiseConvergenceTest(
            atol=1., rtol=1e-7)

    def newDiscreteEq(self, builder):
        return builder.newTransportChargedEquation(
            builder.getMesh(self.mesh), self.name, self.z)

    def load_concentration(self, ctx, eq):
        variables = ctx.varsOf(eq)
        c = variables['x']
        ct = variables['xt']
        variables.update(dict(c=c, ct=ct))
        self.output_concentration(ctx, eq)

    def load(self, ctx, eq):
        super(ConservedSpecies, self).load(ctx, eq)
        if ctx.solver.poissonOnly:
            return
        ctx.varsOf(eq).update(sources=0)
        assert eq.poisson.mesh is eq.mesh, 'multiple meshes currently unsupported'

    def calculate_fluxes(self, ctx, eq, v, D):
        c = ctx.varsOf(eq)['c']
        if not ctx.wants_output:
            return eq.faceflux(c=c, v=v, D=D, full_output=False)
        f = eq.faceflux(c=c, v=v, D=D, full_output=True)
        ctx.outputFace([eq, 'j'], f['flux'], unit=ctx.units.flux)
        ctx.outputFace([eq, 'jdrift'], f['flux_v'],
                       unit=ctx.units.flux)
        ctx.outputFace([eq, 'jdiff'], f['flux_D'], unit=ctx.units.flux)
        ctx.outputFace([eq, 'J'], f['flux']*eq.ze,
                       unit=ctx.units.current_density)
        ctx.outputFace([eq, 'Jdrift'], f['flux_v']*eq.ze,
                       unit=ctx.units.current_density)
        ctx.outputFace([eq, 'Jdiff'], f['flux_D']*eq.ze,
                       unit=ctx.units.current_density)
        return f['flux']

    def evaluate_fluxes(self, ctx, eq):
        raise NotImplementedError()

    def evaluate(self, ctx, eq):
        if ctx.solver.poissonOnly:
            return self.identity(ctx, eq)
        variables = ctx.varsOf(eq)
        j = self.evaluate_fluxes(ctx, eq)
        variables.update(j=j)
        return self.residuals(
            ctx, eq, flux=j, source=variables['sources'], transient=variables['ct'])


class NontransportedSpecies(ConservedSpecies):
    def evaluate_fluxes(self, ctx, eq):
        return None
